import pytest
import os
import shutil
from yadg.extractors.ezchrom.asc import extract as extract_asc
from yadg.extractors.ezchrom.dat import extract as extract_dat
import xarray as xr


@pytest.fixture
def ezchrom_datadir(tmpdir, request):
    filename = request.module.__file__
    test_dir, _ = os.path.splitext(filename)
    base_dir, _ = os.path.split(test_dir)
    for ddir in ["test_x_ezchrom_dat", "test_x_ezchrom_asc"]:
        shutil.copytree(os.path.join(base_dir, ddir), str(tmpdir), dirs_exist_ok=True)
    return tmpdir


@pytest.mark.parametrize(
    "datfile, ascfile",
    [
        ("2023-06-29-007.dat", "2023-06-29-007.dat.asc"),
        ("2023-06-29-014.dat", "2023-06-29-014.dat.asc"),
        ("230324.dat", "230324.dat.asc"),
    ],
)
def test_ezchrom_consistency(datfile, ascfile, ezchrom_datadir):
    os.chdir(ezchrom_datadir)
    aret = extract_dat(fn=datfile, timezone="Europe/Berlin")
    bret = extract_asc(fn=ascfile, timezone="Europe/Berlin", encoding="windows-1252")
    for key in aret.variables:
        try:
            xr.testing.assert_allclose(aret[key], bret[key])
        except AssertionError as e:
            e.args = (e.args[0] + f"Error happened on key: {key!r}\n",)
            raise e



import pytest
import os
import pickle
from yadg.extractors.tomato.json import extract
from .utils import compare_datatrees, datagram_from_file


@pytest.mark.parametrize(
    "infile",
    [
        "MPG2_2022-04-20T213025.275348+0000_data.json",
        "MPG2_2022-04-20T214430.234496+0000_data.json",
    ],
)
def test_tomato_json(infile, datadir):
    os.chdir(datadir)
    ret = extract(fn=infile)
    outfile = f"{infile}.pkl"
    with open(outfile, "rb") as inp:
        ref = pickle.load(inp)
    print(f"{ret=}")
    with open(outfile, "wb") as out:
        pickle.dump(ret, out, 5)
    compare_datatrees(ret, ref, thislevel=True)


@pytest.mark.parametrize(
    "infile",
    [
        "tomato_json_dataschema.1.yml",
        "tomato_json_dataschema.2.yml",
    ],
)
def test_tomato_json_dataschema(infile, datadir):
    os.chdir(datadir)
    ret = datagram_from_file(infile)
    outfile = f"{infile}.pkl"
    with open(outfile, "rb") as inp:
        ref = pickle.load(inp)
    print(f"{ret=}")
    with open(outfile, "wb") as out:
        pickle.dump(ret, out, 5)
    compare_datatrees(ret, ref)



"""
This module includes shared functions for the :mod:`~yadg.extractors.drycal`
extractor, including functions for parsing the files, processing the tabulated data,
and ensuring timestamps are increasing.

.. codeauthor::
    Peter Kraus

"""

from pydantic import BaseModel
from typing import Optional
from xarray import Dataset
import logging
import xarray as xr
from striprtf.striprtf import rtf_to_text

from yadg import dgutils
from yadg.extractors.basic.csv import process_row

logger = logging.getLogger(__name__)


class TimeDate(BaseModel):
    class TimestampSpec(BaseModel, extra="forbid"):
        index: Optional[int] = None
        format: Optional[str] = None

    date: Optional[TimestampSpec] = None
    time: Optional[TimestampSpec] = None


def rtf(
    fn: str,
    encoding: str,
    timezone: str,
) -> Dataset:
    with open(fn, "r", encoding=encoding) as infile:
        rtf = infile.read()
    lines = rtf_to_text(rtf).split("\n")
    for li in range(len(lines)):
        if lines[li].startswith("Sample"):
            si = li
        elif lines[li].startswith("1|"):
            di = li
            break
    # Metadata processing for rtf files is in columns, not rows.
    ml = []
    metadata = dict()
    for line in lines[:si]:
        if line.strip() != "":
            items = [i.strip() for i in line.split("|")]
            if len(items) > 1:
                ml.append(items)
    assert len(ml) == 2 and len(ml[0]) == len(ml[1])
    for i in range(len(ml[0])):
        if ml[0][i] != "":
            metadata[ml[0][i]] = ml[1][i]

    # Process data table
    dl = []
    dl.append(" ".join(lines[si:di]))
    for line in lines[di:]:
        if line.strip() != "":
            dl.append(line)
    headers, units, data = drycal_table(dl, sep="|")
    datecolumns, datefunc, _ = dgutils.infer_timestamp_from(
        spec=TimeDate(time={"index": 4, "format": "%I:%M:%S %p"}), timezone=timezone
    )

    # Process rows
    data_vals = {}
    meta_vals = {"_fn": []}
    for pi, point in enumerate(data):
        vals, devs = process_row(headers[1:], point[1:], datefunc, datecolumns)
        dgutils.append_dicts(vals, devs, data_vals, meta_vals, fn, pi)

    return dgutils.dicts_to_dataset(data_vals, meta_vals, units, False)


def sep(
    fn: str,
    sep: str,
    encoding: str,
    timezone: str,
) -> Dataset:
    with open(fn, "r", encoding=encoding) as infile:
        lines = infile.readlines()
    for li in range(len(lines)):
        if lines[li].startswith("Sample"):
            si = li
        elif lines[li].startswith(f"1{sep}"):
            di = li
            break
    # Metadata processing for csv files is standard.
    metadata = dict()
    for line in lines[:si]:
        if line.strip() != "":
            items = [i.strip() for i in line.split(sep)]
            if len(items) == 2:
                metadata[items[0]] = items[1]

    # Process data table
    dl = list()
    dl.append(" ".join(lines[si:di]))
    for line in lines[di:]:
        if line.strip() != "":
            dl.append(line)
    headers, units, data = drycal_table(dl, sep=sep)

    if "AM" in data[0][-1].upper() or "PM" in data[0][-1].upper():
        fmt = "%I:%M:%S %p"
    else:
        fmt = "%H:%M:%S"
    datecolumns, datefunc, _ = dgutils.infer_timestamp_from(
        spec=TimeDate(time={"index": 4, "format": fmt}), timezone=timezone
    )

    # Process rows
    data_vals = {}
    meta_vals = {"_fn": []}
    for pi, point in enumerate(data):
        vals, devs = process_row(headers[1:], point[1:], datefunc, datecolumns)
        dgutils.append_dicts(vals, devs, data_vals, meta_vals, fn, pi)

    return dgutils.dicts_to_dataset(data_vals, meta_vals, units, False)


def drycal_table(lines: list, sep: str = ",") -> tuple[list, dict, list]:
    """
    DryCal table-processing function.

    Given a table with headers and units in the first line, and data in the following
    lines, this function returns the headers, units, and data extracted from the table.
    The returned values are always of :class:`(str)` type, any post-processing is done
    in the calling routine.
    """
    items = [i.strip() for i in lines[0].split(sep)]
    headers = []
    units = {}
    data = []
    trim = False
    for item in items:
        for rs in [". ", " "]:
            parts = item.split(rs)
            if len(parts) == 2:
                break
        headers.append(parts[0])
        if len(parts) == 2:
            units[parts[0]] = parts[1]
        else:
            units[parts[0]] = " "
    if items[-1] == "":
        trim = True
        headers = headers[:-1]
    for line in lines[1:]:
        cols = line.split(sep)
        assert len(cols) == len(items)
        if trim:
            data.append(cols[:-1])
        else:
            data.append(cols)

    units = dgutils.sanitize_units(units)
    return headers, units, data


def check_timestamps(vals: Dataset) -> Dataset:
    warn = True
    ndays = 0
    utslist = vals.uts.values
    for i in range(1, vals.uts.size):
        if utslist[i] < utslist[i - 1]:
            if warn:
                logger.warning("DryCal log crossing day boundary. Adding offset.")
                warn = False
            uts = utslist[i] + ndays * 86400
            while uts < utslist[i - 1]:
                ndays += 1
                uts = utslist[i] + ndays * 86400
            utslist[i] = uts
    vals["uts"] = xr.DataArray(data=utslist, dims=["uts"])
    vals.attrs["fulldate"] = False
    return vals



import sys
from importlib import metadata
import warnings


def get_yadg_metadata() -> dict:
    """
    Returns current **yadg** metadata.
    """
    md = {
        "yadg_version": metadata.version("yadg"),
        "yadg_command": " ".join(sys.argv),
    }
    return md


def deprecated(arg, depin="4.2", depout="5.0") -> None:
    warnings.simplefilter("always", DeprecationWarning)
    warnings.warn(
        f"'{arg}' has been deprecated in "
        f"yadg-{depin} and will stop working in yadg-{depout}",
        category=DeprecationWarning,
        stacklevel=2,
    )
    warnings.simplefilter("default", DeprecationWarning)



"""
Handles binary files created by EZChrom software.

Usage
`````
Available since ``yadg-5.1``.

.. autopydantic_model:: dgbowl_schemas.yadg.dataschema_5_1.filetype.EZChrom_dat

Schema
``````
.. code-block:: yaml

    datatree.DataTree:
      {{ detector_trace }}:
        coords:
          uts:            !!float               # Unix timestamp
          elution_time:   !!float               # Elution time
        data_vars:
          signal:         (uts, elution_time)   # Signal data

Metadata
````````
No metadata is currently extracted. If you need some particular metadata, please open
an issue.

Notes on file structure
```````````````````````
Data in these files is stored as an OLE file, which is first processed using the
:mod:`olefile` library.

The timestamp is stored as an OLE timestamp in the ``Chrom Header`` stream.

The metadata for each trace are stored within the ``Detector Trace Handler`` stream,
and contain the X- and Y-axis multiplier, Y-axis units, and some other metadata.

The data for each trace are stored within the ``Detector Data`` "directory" within the
OLE file, with one stream per trace.

Uncertainties
`````````````
The uncertainties in ``signal`` as well as ``elution_time`` are set to the axis
multiplier.


.. codeauthor::
    Peter Kraus

"""

import olefile
import xarray as xr
from datatree import DataTree
import numpy as np

from yadg import dgutils


detector_trace_struct = [
    ("trace_name", 22, "pascal"),
    ("position", 4, "pascal"),
    ("x_mul", 0, "f4"),
    ("y_unit", 0, "pascal"),
    ("y_mul", 0, "f4"),
    ("tst", 33, "pascal"),
    ("y_unit2", 0, "pascal"),
    ("time", 21, "f4"),
]


def extract(
    *,
    fn: str,
    timezone: str,
    **kwargs: dict,
) -> DataTree:
    # Read data from the OLE file
    dd = {}
    with olefile.OleFileIO(fn) as of:
        ch = of.openstream(["Chrom Header"]).read()
        dth = of.openstream(["Detector Trace Handler"]).read()
        for path in of.listdir():
            if len(path) == 2 and path[0] == "Detector Data":
                dd[path[1]] = of.openstream(path).read()
    # Timestamp
    ole_timestamp = dgutils.read_value(data=ch, offset=8, dtype="f8")
    uts = dgutils.ole_to_uts(ole_timestamp, timezone)

    # Trace metadata
    offset = 31
    dtp = {}
    for _ in dd:
        params = {}
        for name, delta, dtype in detector_trace_struct:
            offset += delta
            params[name] = dgutils.read_value(data=dth, offset=offset, dtype=dtype)
            if dtype == "pascal":
                offset += dgutils.read_value(data=dth, offset=offset, dtype="u1") + 1
            elif dtype == "f4":
                offset += 4
        dtp[f"Detector {params['trace_name']} Trace"] = params

    # Trace data
    dt = {}
    for key, vals in dd.items():
        par = dtp[key]
        npoints = dgutils.read_value(data=vals, offset=4, dtype="u4")
        yvals = np.frombuffer(vals, offset=20, count=npoints, dtype="i4") * par["y_mul"]
        ydevs = np.ones(npoints) * par["y_mul"]
        yunits = {"units": par["y_unit"].replace("25", "").strip()}
        xvals = np.arange(0, npoints) * par["x_mul"]
        xdevs = np.ones(npoints) * par["x_mul"]
        xunits = {"units": "s"}
        ds = xr.Dataset(
            data_vars={
                "signal": (["uts", "elution_time"], [yvals], yunits),
                "signal_std_err": (["uts", "elution_time"], [ydevs], yunits),
                "elution_time_std_err": (["elution_time"], xdevs, xunits),
            },
            coords={
                "elution_time": (["elution_time"], xvals, xunits),
                "uts": (["uts"], [uts]),
            },
        )
        for var in ds.variables:
            if f"{var}_std_err" in ds.variables:
                ds[var].attrs["ancillary_variables"] = f"{var}_std_err"
            elif var.endswith("_std_err"):
                end = var.index("_std_err")
                if var[:end] in ds.variables:
                    ds[var].attrs["standard_name"] = f"{var[:end]} standard_error"
        dt[f"/{key}"] = ds
    return DataTree.from_dict(dt)



import re


def panalytical_comment(line: str) -> dict:
    """Processes a comments from the file header into a dictionary.

    Parameters
    ----------
    line
        A line containing the comment.

    Returns
    -------
    dict
        A dictionary containing the processed comment.

    """

    if line.startswith("Configuration="):
        split = [s.split("=") for s in line.split(", ")]
        __, values = list(zip(*split))
        keys = ["configuration", "owner", "creation_date"]
    elif line.startswith("Goniometer="):
        split = [s.replace("=", ":").split(":") for s in line.split(";")]
        __, values = list(zip(*split))
        keys = ["goniometer", "min_step_size_2theta", "min_step_size_omega"]
    elif line.startswith("Sample stage="):
        __, values = line.split("=")
        keys = ["sample_stage"]
    elif line.startswith("Diffractometer system="):
        __, values = line.split("=")
        keys = ["diffractometer_system"]
    elif line.startswith("Measurement program="):
        split = [s.split("=") for s in line.split(", ")]
        __, values = list(zip(*split))
        keys = ["measurement_program", "identifier"]
    elif line.startswith("Fine Calibration Offset for 2Theta"):
        __, values = line.split(" = ")
        keys = ["calib_offset_2theta"]
    values = [values] if isinstance(values, str) else values
    return dict(zip(keys, values))


def snake_case(s: str) -> str:
    """Converts Sentence case. and camelCase strings to snake_case.

    From https://stackoverflow.com/a/1176023

    Parameters
    ----------
    s
        The input string to be converted.

    Returns
    -------
    str
        The corresponding snake_case string.

    """
    s = "".join([s.capitalize() for s in s.replace(".", "").split()])
    s = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", s)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", s).lower()



import pytest
import os
import pickle
from yadg.extractors.touchstone.snp import extract
from .utils import compare_datatrees


@pytest.mark.parametrize(
    "infile, locale",
    [
        ("picovna.s1p", "en_GB"),
        ("Device_r_40um.s1p", "en_GB"),
        ("Device_r_60um.s1p", "en_GB"),
        ("Fig8_0.6cm.s1p", "en_GB"),
        ("Fig8_5.6cm.s1p", "en_GB"),
        ("SCENARIO1_pulserside_30k_3G.s1p", "en_GB"),
        ("SCENARIO4_pulserside_30k_3G.s1p", "en_GB"),
        ("CABLE_3_5MM_1N_TYPE_CONNECTORS.S2P", "en_GB"),
        ("VNA_radial_middle.s1p", "en_GB"),
        ("VNA_radial_middle.s2p", "en_GB"),
        ("2024-06-20-01-empty.s1p", "de_DE"),
    ],
)
def test_touchstone_snp(infile, locale, datadir):
    os.chdir(datadir)
    ret = extract(fn=infile, encoding="utf-8", timezone="Europe/Berlin", locale=locale)
    outfile = f"{infile}.pkl"
    with open(outfile, "rb") as inp:
        ref = pickle.load(inp)
    print(f"{ret=}")
    with open(outfile, "wb") as out:
        pickle.dump(ret, out, 5)
    compare_datatrees(ret, ref, thislevel=True)



import pytest
import os
import shutil
from yadg.extractors.empalc.xlsx import extract as extract_xls
from yadg.extractors.empalc.csv import extract as extract_csv
import xarray as xr


@pytest.fixture
def _datadir(tmpdir, request):
    filename = request.module.__file__
    test_dir, _ = os.path.splitext(filename)
    base_dir, _ = os.path.split(test_dir)
    for ddir in ["test_x_empalc_xlsx", "test_x_empalc_csv"]:
        shutil.copytree(os.path.join(base_dir, ddir), str(tmpdir), dirs_exist_ok=True)
    return tmpdir


@pytest.mark.parametrize(
    "afile, bfile",
    [("Cu-25p_v2.xlsx", "Cu-25p_v2.csv")],
)
def test_empalc_consistency(afile, bfile, _datadir):
    os.chdir(_datadir)
    aret = extract_xls(fn=afile)
    bret = extract_csv(fn=bfile, encoding="utf-8")

    for key in aret.variables:
        if key.endswith("std_err"):
            continue
        try:
            xr.testing.assert_allclose(aret[key], bret[key])
        except AssertionError as e:
            e.args = (e.args[0] + f"Error happened on key: {key!r}\n",)
            raise e



"""
This is an example extractor, used mainly for testing of the :mod:`yadg` package.
It provides no real functionality.

Usage
`````
Available since ``yadg-4.0``.

.. autopydantic_model:: dgbowl_schemas.yadg.dataschema_5_1.filetype.Example


Schema
``````
The output schema is only defined for the ``tomato.json`` filetype.

.. code-block:: yaml

  datatree.DataTree:
    coords:
      uts:              !!float      # The current timestamp
    data_vars:
      {{ param_keys }}  (None)       # All parameter key/value pairs

Metadata
````````
No metadata is returned.

.. codeauthor::
    Peter Kraus

"""

from pydantic import BaseModel
from yadg import dgutils
from datatree import DataTree


def extract(
    *,
    fn: str,
    parameters: BaseModel,
    **kwargs: dict,
) -> DataTree:
    kwargs = {} if parameters is None else parameters.model_dump()
    if "parser" in kwargs:
        del kwargs["parser"]
    data_vals = {k: [v] for k, v in kwargs.items()}
    data_vals["uts"] = [dgutils.now()]
    meta_vals = {}
    return DataTree(dgutils.dicts_to_dataset(data_vals, meta_vals, fulldate=False))



"""
Handles processing of the csv version of the structured format produced by Agilent's
Online LC device at Empa. It contains three sections:

  - metadata section,
  - table containing sampling information,
  - table containing analysed chromatography data.

Usage
`````
Available since ``yadg-4.2``.

.. autopydantic_model:: dgbowl_schemas.yadg.dataschema_5_1.filetype.EmpaLC_csv

Schema
``````
.. code-block:: yaml

    datatree.DataTree:
      coords:
        uts:            !!float               # Unix timestamp
        species:        !!str                 # Species name
      data_vars:
        height:         (uts, species)        # Peak height
        area:           (uts, species)        # Integrated peak area
        concentration:  (uts, species)        # Peak area with calibration applied
        retention time: (uts, species)        # Position of peak maximum

Metadata
````````
The following metadata is extracted:

  - ``sequence``: Sample / sequence name.
  - ``description``: A free-form description of the experiment.
  - ``username``: User name used to generate the file.
  - ``datafile``: Original path of the result file.
  - ``version``: Version of the export function used to generate the file.

.. codeauthor::
    Peter Kraus

"""

import logging
import datetime
from uncertainties.core import str_to_number_with_uncert as tuple_fromstr
import xarray as xr
import numpy as np
from datatree import DataTree

logger = logging.getLogger(__name__)


def extract(
    *,
    fn: str,
    encoding: str,
    **kwargs: dict,
) -> DataTree:
    with open(fn, "r", encoding=encoding, errors="ignore") as infile:
        lines = infile.readlines()

    metadata = {}
    while len(lines) > 0:
        line = lines.pop(0)
        if len(lines) == 0:
            raise RuntimeError(
                f"Last line of file '{fn}' read during metadata section."
            )
        elif line.strip() == "":
            break
        elif line.strip().startswith("Sequence name"):
            metadata["sequence"] = line.split(":,")[1].strip()
        elif line.strip().startswith("Description"):
            metadata["description"] = line.split(":,")[1].strip()
        elif line.strip().startswith("Acquired by"):
            metadata["username"] = line.split(":,")[1].strip()
        elif line.strip().startswith("Data path"):
            metadata["datafile"] = line.split(":,")[1].strip()
        elif line.strip().startswith("Report version"):
            metadata["version"] = int(line.split(":,")[1])

    if metadata.get("version", None) is None:
        raise RuntimeError(f"Report version in file '{fn}' was not specified.")

    samples = {}
    while len(lines) > 0:
        line = lines.pop(0)
        if len(lines) == 0:
            raise RuntimeError(f"Last line of file '{fn}' read during samples section.")
        elif line.strip() == "":
            break
        elif "Line#" in line:
            headers = [i.strip() for i in line.split(",")]
        else:
            data = [i.strip() for i in line.split(",")]
            sample = {
                "location": data[headers.index("Location")],
                "injection date": data[headers.index("Injection Date")],
                "acquisition": {
                    "method": data[headers.index("Acq Method Name")],
                    "version": data[headers.index("Acq Method Version")],
                },
                "integration": {
                    "method": data[headers.index("Injection DA Method Name")],
                    "version": data[headers.index("Injection DA Method Version")],
                },
                "offset": data[headers.index("Time offset")],
            }
            if sample["offset"] != "":
                sn = data[headers.index("Sample Name")]
                samples[sn] = sample

    svals = samples.values()
    if len(svals) == 0:
        raise RuntimeError(
            f"No complete sample data found in file '{fn}'. "
            "Have you added time offsets?"
        )
    r = next(iter(svals))
    # check that acquisition and integration methods are consistent throughout file:
    if any([s["acquisition"]["method"] != r["acquisition"]["method"] for s in svals]):
        logger.warning("Acquisition method is inconsistent in file '%s'.", fn)
    if any([s["integration"]["method"] != r["integration"]["method"] for s in svals]):
        logger.warning("Integration method is inconsistent in file '%s'.", fn)

    metadata["method"] = r["acquisition"]["method"]

    species = set()
    while len(lines) > 0:
        line = lines.pop(0)
        if len(lines) == 0:
            break
        elif line.strip() == "":
            break
        elif "Line#" in line:
            headers = [i.strip() for i in line.split(",")]
        else:
            data = [i.strip() for i in line.split(",")]
            sn = data[headers.index("Sample Name")]
            cn = data[headers.index("Compound")]
            species.add(cn)

            h = data[headers.index("Peak Height")]
            if h != "":
                if "height" not in samples[sn]:
                    samples[sn]["height"] = {}
                samples[sn]["height"][cn] = tuple_fromstr(h)

            A = data[headers.index("Area")]
            if A != "":
                if "area" not in samples[sn]:
                    samples[sn]["area"] = {}
                samples[sn]["area"][cn] = tuple_fromstr(A)

            if metadata["version"] == 2:
                c = data[headers.index("Concentration")]
            else:
                logger.warning(
                    "Report version '%d' in file '%s' not understood.",
                    metadata["version"],
                    fn,
                )
                c = data[headers.index("Concentration")]
            if c != "":
                if "concentration" not in samples[sn]:
                    samples[sn]["concentration"] = {}
                samples[sn]["concentration"][cn] = tuple_fromstr(c)

            rt = data[headers.index("RT [min]")]
            if rt != "":
                if "retention time" not in samples[sn]:
                    samples[sn]["retention time"] = {}
                samples[sn]["retention time"][cn] = tuple_fromstr(rt)
    units = {
        "height": None,
        "area": None,
        "concentration": "mmol/l",
        "retention time": "min",
    }
    species = sorted(species)
    data = []
    for k, v in samples.items():
        # Remove unnecessary parameters
        del v["acquisition"]
        del v["integration"]
        v["sampleid"] = k
        # Process offset to uts
        offset = v.pop("offset")
        t = None
        for fmt in {"%H:%M:%S"}:
            try:
                t = datetime.datetime.strptime(offset, fmt)
            except ValueError:
                continue
        if t is None:
            try:
                td = datetime.timedelta(minutes=float(offset))
            except ValueError:
                raise RuntimeError(
                    f"It was not possible to parse offset '{offset}' present in file "
                    f"'{fn}' using known formats."
                )
        else:
            td = datetime.timedelta(hours=t.hour, minutes=t.minute, seconds=t.second)
        point = {"uts": td.total_seconds()}
        vals = {}
        devs = {}
        for kk in {"height", "area", "concentration", "retention time"}:
            val = v.get(kk, {})
            vals[kk], devs[kk] = zip(*[val.get(cn, (np.nan, np.nan)) for cn in species])
        point["vals"] = vals
        point["devs"] = devs
        data.append(point)

    data_vars = {}
    for kk in {"height", "area", "concentration", "retention time"}:
        data_vars[kk] = (
            ["uts", "species"],
            [i["vals"][kk] for i in data],
            {"anciliary_variables": f"{kk}_std_err"},
        )
        data_vars[f"{kk}_std_err"] = (
            ["uts", "species"],
            [i["devs"][kk] for i in data],
            {"standard_name": f"{kk} standard_error"},
        )
        if units[kk] is not None:
            data_vars[kk][2]["units"] = units[kk]
            data_vars[f"{kk}_std_err"][2]["units"] = units[kk]

    ds = xr.Dataset(
        data_vars=data_vars,
        coords={
            "species": (["species"], species),
            "uts": (["uts"], [i["uts"] for i in data]),
        },
        attrs=dict(original_metadata=metadata),
    )

    return DataTree(ds)



"""
Used to process files generated using Agilent PNA-L N5320C via its LabVIEW driver.
This file format includes a header, with the values of bandwidth and averaging,
and three tab-separated columns containing the frequency :math:`f`, and the real
and imaginary parts of the complex reflection coefficient :math:`\\Gamma(f)`.

Note that no timestamps are present in the file and have to be supplied externally,
e.g. from the file name. One trace per file. As the MCPT set-up for which this
extractor was designed always uses the ``S11`` port, the node name is is hard-coded to
this value.

Usage
`````
Available since ``yadg-3.0``.

.. autopydantic_model:: dgbowl_schemas.yadg.dataschema_5_1.filetype.FHI_vna

Schema
``````
.. code-block:: yaml

    datatree.DataTree:
      S11:              !!xarray.Dataset
        coords:
            freq:       !!float     # An array of measurement frequencies
        data_vars:
            Re(G):      (freq)      # Real part of Γ
            Im(G):      (freq)      # Imaginary part of Γ
            average:    (None)      # Number of traces averaged
            bandwidth:  (None)      # Filter bandwidth

Metadata
````````
No metadata is returned.

.. codeauthor::
    Peter Kraus

"""

from uncertainties.core import str_to_number_with_uncert as tuple_fromstr
import xarray as xr
from datatree import DataTree


def extract(
    *,
    fn: str,
    encoding: str,
    **kwargs: dict,
) -> DataTree:
    with open(fn, "r", encoding=encoding) as infile:
        lines = infile.readlines()
    assert (
        len(lines) > 2
    ), f"qftrace: Only {len(lines)-1} points supplied in {fn}; fitting impossible."

    # process header
    bw = [10000.0, 1.0]
    avg = 15
    if ";" in lines[0]:
        items = lines.pop(0).split(";")
        for item in items:
            if item.startswith("BW"):
                bw = tuple_fromstr(item.split("=")[-1].strip())
            if item.startswith("AVG"):
                avg = int(item.split("=")[-1].strip())
    fsbw = bw[0] / avg

    # calculate precision of trace
    freq = {"vals": [], "devs": []}
    real = {"vals": [], "devs": []}
    imag = {"vals": [], "devs": []}
    for line in lines:
        f, re, im = line.strip().split()
        fn, fs = tuple_fromstr(f)
        fs = max(fs, fsbw)
        ren, res = tuple_fromstr(re)
        imn, ims = tuple_fromstr(im)
        freq["vals"].append(fn)
        freq["devs"].append(fs)
        real["vals"].append(ren)
        real["devs"].append(res)
        imag["vals"].append(imn)
        imag["devs"].append(ims)

    vals = xr.Dataset(
        data_vars={
            "Re(G)": (
                ["freq"],
                real["vals"],
                {"ancillary_variables": "Re(G)_std_err"},
            ),
            "Re(G)_std_err": (
                ["freq"],
                real["devs"],
                {"standard_name": "Re(G) standard_error"},
            ),
            "Im(G)": (
                ["freq"],
                imag["vals"],
                {"ancillary_variables": "Im(G)_std_err"},
            ),
            "Im(G)_std_err": (
                ["freq"],
                imag["devs"],
                {"standard_name": "Im(G) standard_error"},
            ),
            "average": avg,
            "bandwidth": (
                [],
                bw[0],
                {"units": "Hz", "ancillary_variables": "bandwidth_std_err"},
            ),
            "bandwidth_std_err": (
                [],
                bw[1],
                {"units": "Hz", "standard_name": "bandwidth standard_error"},
            ),
        },
        coords={
            "freq": (
                ["freq"],
                freq["vals"],
                {"units": "Hz", "ancillary_variables": "freq_std_err"},
            ),
            "freq_std_err": (
                ["freq"],
                freq["devs"],
                {"units": "Hz", "standard_name": "freq standard_error"},
            ),
        },
    )

    return DataTree.from_dict(dict(S11=vals))



import pytest
import os
import pickle
from yadg.extractors.drycal.txt import extract
from .utils import compare_datatrees


@pytest.mark.parametrize(
    "infile",
    [
        "20211011_DryCal_out.txt",
    ],
)
def test_drycal_txt(infile, datadir):
    os.chdir(datadir)
    ret = extract(fn=infile, encoding="utf8", timezone="Europe/Berlin")
    outfile = f"{infile}.pkl"
    with open(outfile, "rb") as inp:
        ref = pickle.load(inp)
    print(f"{ret=}")
    with open(outfile, "wb") as out:
        pickle.dump(ret, out, 5)
    compare_datatrees(ret, ref, thislevel=True)



"""
Handles processing of xy exports of Panalytical XRD files. When possible, the xrdml or
csv files should be used instead.

Usage
`````
Available since ``yadg-4.2``.

.. autopydantic_model:: dgbowl_schemas.yadg.dataschema_5_1.filetype.Panalytical_xy

Schema
``````
.. code-block:: yaml

    datatree.DataTree:
      coords:
        angle:          !!float               # 2θ angle
      data_vars:
        intensity:      (angle)          # Measured intensity

Metadata
````````
No metadata is present in files.

Notes on file structure
```````````````````````

These files basically just contain the ``[Scan points]`` part of Panalytical csv files.
As a consequence, no metadata is recorded, and the format does not have an associated
timestamp.

.. codeauthor::
    Nicolas Vetsch,
    Peter Kraus
"""

from uncertainties.core import str_to_number_with_uncert as tuple_fromstr
import numpy as np
from datatree import DataTree
import xarray as xr


def extract(
    *,
    fn: str,
    encoding: str,
    **kwargs: dict,
) -> DataTree:
    with open(fn, "r", encoding=encoding) as xy_file:
        xy = xy_file.readlines()
    datapoints = [li.strip().split() for li in xy]
    angle, intensity = list(zip(*datapoints))
    angle, _ = list(zip(*[tuple_fromstr(a) for a in angle]))
    insty, _ = list(zip(*[tuple_fromstr(i) for i in intensity]))
    idevs = np.ones(len(insty))
    adiff = np.abs(np.diff(angle)) * 0.5
    adiff = np.append(adiff, adiff[-1])
    vals = xr.Dataset(
        data_vars={
            "intensity": (
                ["angle"],
                list(insty),
                {"units": "counts", "ancillary_variables": "intensity_std_err"},
            ),
            "intensity_std_err": (
                ["angle"],
                idevs,
                {"units": "counts", "standard_name": "intensity standard_error"},
            ),
            "angle_std_err": (
                ["angle"],
                adiff,
                {"units": "deg", "standard_name": "angle standard_error"},
            ),
        },
        coords={
            "angle": (
                ["angle"],
                list(angle),
                {"units": "deg", "ancillary_variables": "angle_std_err"},
            ),
        },
    )
    return DataTree(vals)



"""
Handles the reading and processing of volumetric flow meter data exported from the
MesaLabs DryCal software as a csv file.

.. note::

    The date information is missing in the timestamps of the exported files and has to
    be supplied externally.

Usage
`````
Available since ``yadg-4.0``.

.. autopydantic_model:: dgbowl_schemas.yadg.dataschema_5_1.filetype.Drycal_csv

Schema
``````
.. code-block:: yaml

    datatree.DataTree:
      coords:
        uts:            !!float               # Unix timestamp, without date
      data_vars:
        DryCal:         (uts)                 # Standardised flow rate
        DryCal Avg.:    (uts)                 # Running average of the flow rate
        Temp.:          (uts)                 # Measured flow temperature
        Pressure:       (uts)                 # Measured flow pressure

Metadata
````````
The following metadata is extracted:

    - ``product``: Model name of the MesaLabs device.
    - ``serial number``: Serial number of the MesaLabs device.

Uncertainties
`````````````
All uncertainties are derived from the string representation of the floats.

.. codeauthor::
    Peter Kraus

"""

from datatree import DataTree
from yadg.extractors.drycal import common


def extract(
    *,
    fn: str,
    encoding: str,
    timezone: str,
    **kwargs: dict,
) -> DataTree:
    vals = common.sep(fn, ",", encoding, timezone)
    return DataTree(common.check_timestamps(vals))



"""
For processing Inficon Fusion zipped data. This is a wrapper parser which unzips the
provided zip file, and then uses the :mod:`yadg.extractors.fusion.json` extractor
to parse every fusion-data file present in the archive.

Contains both the data from the raw chromatogram and the post-processed results.

Usage
`````
Available since ``yadg-4.0``.

.. autopydantic_model:: dgbowl_schemas.yadg.dataschema_5_1.filetype.Fusion_zip

Schema
``````
.. code-block:: yaml

    datatree.DataTree:
      coords:
        uts:              !!float
        species:          !!str
      data_vars:
        height:           (uts, species)        # Peak height at maximum
        area:             (uts, species)        # Integrated peak area
        concentration:    (uts, species)        # Calibrated peak area
        xout:             (uts, species)        # Mole fraction (normalized conc.)
        retention time:   (uts, species)        # Peak retention time
      {{ detector_name }}:
        coords:
          uts:            !!float               # Unix timestamp
          elution_time:   !!float               # Elution time
        data_vars:
          signal:         (uts, elution_time)   # Signal data
          valve:          (uts)                 # Valve position

Metadata
````````
No metadata is currently extracted.

.. codeauthor::
    Peter Kraus

"""

import zipfile
import tempfile
import os
from datatree import DataTree

from yadg.extractors.fusion.json import extract as extract_json
from yadg import dgutils


def extract(
    *,
    fn: str,
    timezone: str,
    encoding: str,
    **kwargs: dict,
) -> DataTree:
    zf = zipfile.ZipFile(fn)
    with tempfile.TemporaryDirectory() as tempdir:
        zf.extractall(tempdir)
        dt = None
        filenames = [ffn for ffn in os.listdir(tempdir) if ffn.endswith("fusion-data")]
        for ffn in sorted(filenames):
            path = os.path.join(tempdir, ffn)
            fdt = extract_json(fn=path, timezone=timezone, encoding=encoding, **kwargs)
            dt = dgutils.merge_dicttrees(dt, fdt.to_dict(), "identical")
    return DataTree.from_dict(dt)



import numpy as np
from typing import Union, Any


def read_pascal_string(pascal_bytes: bytes, encoding: str = "windows-1252") -> str:
    """Parses a length-prefixed string given some encoding.

    Parameters
    ----------
    bytes
        The bytes of the string starting at the length-prefix byte.

    encoding
        The encoding of the string to be converted.

    Returns
    -------
    str
        The string decoded from the input bytes.

    """
    mul = 2 if encoding in {"utf-16"} else 1
    lstr = int.from_bytes(pascal_bytes[0:1], byteorder="big") * mul
    if len(pascal_bytes) < lstr + 1:
        raise ValueError("Insufficient number of bytes.")
    string_bytes = pascal_bytes[1 : lstr + 1]
    return string_bytes.decode(encoding)


def read_value(
    data: bytes,
    offset: int,
    dtype: Union[np.dtype, str],
    encoding: str = "windows-1252",
) -> Any:
    """Reads a single value or a set of values from a buffer at a certain offset.

    Just a handy wrapper for np.frombuffer(..., count=1) With the added benefit of
    allowing the 'pascal' keyword as an indicator for a length-prefixed
    string.

    The read value is converted to a built-in datatype using
    np.dtype.item().

    Parameters
    ----------
    data
        An object that exposes the buffer interface. Here always bytes.

    offset
        Start reading the buffer from this offset (in bytes).

    dtype
        Data-type to read in.

    encoding
        The encoding of the bytes to be converted.

    Returns
    -------
    Any
        The unpacked and converted value from the buffer.

    """
    if dtype == "pascal":
        # Allow the use of 'pascal' in all of the dtype maps.
        return read_pascal_string(data[offset:], encoding="windows-1252")
    elif dtype in {"utf-8", "utf-16"}:
        return read_pascal_string(data[offset:], encoding=dtype)
    value = np.frombuffer(data, offset=offset, dtype=dtype, count=1)
    item = value.item()
    if value.dtype.names:
        item = [i.decode(encoding) if isinstance(i, bytes) else i for i in item]
        return dict(zip(value.dtype.names, item))
    return item.decode(encoding) if isinstance(item, bytes) else item



"""
For processing of BioLogic's EC-Lab binary modular files.

Usage
`````
Available since ``yadg-4.0``.

.. autopydantic_model:: dgbowl_schemas.yadg.dataschema_5_1.filetype.EClab_mpt

Schema
``````
The ``.mpt`` files contain many columns that vary depending on the electrochemical
technique used. Below is shown a list of columns that can be expected to be present
in a typical ``.mpt`` file.

.. code-block:: yaml

    datatree.DataTree:
      coords:
        uts:            !!float     # Unix timestamp, without date
      data_vars:
        Ewe             (uts)       # Potential of the working electrode
        Ece             (uts)       # Potential of the counter electrode, if present
        I               (uts)       # Instantaneous current
        time            (uts)       # Time elapsed since the start of the experiment
        <Ewe>           (uts)       # Average Ewe potential since last data point
        <Ece>           (uts)       # Average Ece potential since last data point
        <I>             (uts)       # Average current since last data point
        ...

.. note::

     Note that in most cases, either the instantaneous or the averaged quantities are
     stored - only rarely are both available!

Notes on file structure
```````````````````````
These human-readable files are sectioned into headerlines and datalines.
The header part of the ``.mpt`` files is made up of information that can be found
in the settings, log and loop modules of the binary ``.mpr`` file.

If no header is present, the timestamps will instead be calculated from
the file's ``mtime()``.

Metadata
````````
The metadata will contain the information from the header of the file.

.. note ::

    The mapping between metadata parameters between ``.mpr`` and ``.mpt`` files
    is not yet complete.

.. codeauthor::
    Nicolas Vetsch,
    Peter Kraus

"""

import logging
from typing import Any
from babel.numbers import parse_decimal
from datatree import DataTree
from yadg import dgutils
from .common.techniques import get_devs, param_from_key
from .common.mpt_columns import column_units

logger = logging.getLogger(__name__)


def process_settings(lines: list[str]) -> dict[str, str]:
    settings = {}
    for line in lines:
        items = [item.strip() for item in line.split(":")]
        if len(items) > 2:
            items = [items[0], ":".join(items[1:])]
        if len(items) == 2 and items[1] != "":
            settings[items[0]] = items[1]
    return settings


def process_params(technique: str, lines: list[str], locale: str) -> dict[str, Any]:
    params = {}
    if len(lines) == 0:
        logger.warning("No params section was found.")
        return params
    # The sequence param columns are always allocated 20 characters.
    n_sequences = int(len(lines[0]) / 20)
    prev = None
    for line in lines:
        items = [line[seq * 20 : (seq + 1) * 20].strip() for seq in range(n_sequences)]
        try:
            vals = [float(parse_decimal(val, locale=locale)) for val in items[1:]]
        except ValueError:
            vals = items[1:]
        if items[0] in {"vs."}:
            name = f"{prev} {items[0]}"
        else:
            name = items[0]
        if name not in params:
            params[name] = vals
        else:
            raise RuntimeError(f"Trying to assing same parameter {items[0]!r} twice.")
        prev = name
    return params


def process_header(
    lines: list[str],
    timezone: str,
    locale: str,
) -> tuple[dict, list, dict]:
    """Processes the header lines.

    Parameters
    ----------
    lines
        The header lines, starting at line 3 (which is an empty line),
        right after the `"Nb header lines : "` line.

    Returns
    -------
    tuple[dict, dict]
        A dictionary containing the settings (and the technique
        parameters) and a dictionary containing the loop indexes.

    """
    sections = "\n".join(lines).split("\n\n")
    # Can happen that no settings are present but just a loops section.
    assert not sections[1].startswith("Number of loops : "), "no settings present"
    # Again, we need the acquisition time to get timestamped data.
    assert len(sections) >= 3, "no settings present"
    technique = sections[1].strip()

    lines = sections[2].split("\n")
    for li, line in enumerate(lines):
        if line.startswith("Cycle Definition :"):
            break

    settings = process_settings(lines[:li])
    params = process_params(technique, lines[li + 1 :], locale)

    # Parse the acquisition timestamp.
    if "Acquisition started on" in settings:
        timestamp = settings["Acquisition started on"]
        for fmt in (
            "%m/%d/%Y %H:%M:%S",
            "%m.%d.%Y %H:%M:%S",
            "%m/%d/%Y %H:%M:%S.%f",
            "%m-%d-%Y %H:%M:%S",
        ):
            uts = dgutils.str_to_uts(
                timestamp=timestamp, format=fmt, timezone=timezone, strict=False
            )
            if uts is not None:
                break
        else:
            raise NotImplementedError(f"Time format for {timestamp} not implemented.")
    else:
        uts = None

    loops = None
    if len(sections) >= 4 and sections[-1].startswith("Number of loops : "):
        # The header contains a loops section.
        loops_lines = sections[-1].split("\n")
        n_loops = int(loops_lines[0].split(":")[-1])
        indexes = []
        for n in range(n_loops):
            index = loops_lines[n + 1].split("to")[0].split()[-1]
            indexes.append(int(index))
        loops = {"n_loops": n_loops, "indexes": indexes}
    ret = {
        "uts": uts,
        "technique": technique,
        "settings": settings,
        "params": params,
        "loops": loops,
    }
    return ret


def process_data(
    lines: list[str],
    Eranges: list[float],
    Iranges: list[float],
    controls: list[str],
    locale: str,
):
    """Processes the data lines.

    Parameters
    ----------
    lines
        The data lines, starting right after the last header section.
        The first line is an empty line, the column names can be found
        on the second line.

    Returns
    -------
    dict
        A dictionary containing the datapoints in the format
        ([{column -> value}, ..., {column -> value}]). If the column
        unit is set to None, the value is an int. Otherwise, the value
        is a dict with value ("n"), sigma ("s"), and unit ("u").

    """
    # At this point the first two lines have already been read.
    # Remove extra column due to an extra tab in .mpt file column names.
    names = lines[1].split("\t")[:-1]
    units = dict()
    columns = list()
    for n in names:
        c, u = column_units[n.strip()]
        if c in columns:
            logger.warning("Duplicate column '%s' with unit '%s'.", c, u)
            c = f"duplicate {c}"
        columns.append(c)
        if u is not None:
            units[c] = u
    # Remove empty lines from data_lines, see issue #151.
    data_lines = [line for line in lines[2:] if line.strip() != ""]
    allvals = dict()
    allmeta = dict()
    warn_I_range = False
    for li, line in enumerate(data_lines):
        values = line.split("\t")
        vals = dict()
        devs = dict()
        for ci, name in enumerate(columns):
            value = values[ci]
            if units.get(name) is None:
                ival = int(parse_decimal(value, locale=locale))
                if name == "I Range":
                    vals[name] = param_from_key("I_range", ival)
                else:
                    vals[name] = ival
            else:
                try:
                    dec = parse_decimal(value, locale=locale)
                    vals[name] = float(dec)
                    exp = dec.as_tuple().exponent
                    devs[name] = float("nan") if isinstance(exp, str) else 10**exp
                except ValueError:
                    sval = value.strip()
                    vals[name] = sval

        Ns = vals.get("Ns", 0)
        Erange = Eranges[Ns] if isinstance(Eranges, list) else Eranges
        Irstr = Iranges[Ns] if isinstance(Iranges, list) else Iranges
        if "I Range" in vals:
            Irstr = vals["I Range"]
        Irange = param_from_key("I_range", Irstr, to_str=False)

        # I Range can be None if it's set to "Auto", "PAC" or other such string.
        if Irange is None:
            warn_I_range = True
            Irange = 1.0

        if "control_VI" in vals:
            icv = controls[Ns] if isinstance(controls, list) else controls
            name = f"control_{icv}"
            vals[name] = vals.pop("control_VI")
            units[name] = "mA" if icv in {"I", "C"} else "V"
        devs = get_devs(vals=vals, units=units, Erange=Erange, Irange=Irange, devs=devs)
        dgutils.append_dicts(vals, devs, allvals, allmeta, li=li)
    if warn_I_range:
        logger.warning("I Range could not be understood, defaulting to 1 A.")

    ds = dgutils.dicts_to_dataset(allvals, allmeta, units, fulldate=False)
    return ds


def extract(
    *,
    fn: str,
    encoding: str,
    locale: str,
    timezone: str,
    **kwargs: dict,
) -> DataTree:
    file_magic = "EC-Lab ASCII FILE\n"
    with open(fn, "r", encoding=encoding) as mpt_file:
        assert mpt_file.read(len(file_magic)) == file_magic, "invalid file magic"
        mpt = mpt_file.read()
    lines = mpt.split("\n")
    nb_header_lines = int(lines[0].split()[-1])
    header_lines = lines[: nb_header_lines - 3]
    data_lines = lines[nb_header_lines - 3 :]
    settings, params = {}, []

    if nb_header_lines <= 3:
        logger.warning("Header contains no settings and hence no timestamp.")
        start_time = 0.0
        fulldate = False
        Eranges = 10.0
        logger.warning("E Range not specified due to missing header, setting to 10 V.")
        Iranges = "1 A"
        logger.warning("I Range not specified due to missing header, setting to 1 A.")
        controls = None
    else:
        header = process_header(header_lines, timezone, locale)
        start_time = header.get("uts")
        settings = header.get("settings")
        params = header.get("params")
        fulldate = True
        Er_max = params.get("E range max (V)", [10.0])
        Er_min = params.get("E range min (V)", [0.0])
        Eranges = [_max - _min for _max, _min in zip(Er_max, Er_min)]
        Iranges = params.get("I Range", ["1 A"])
        controls = params.get("Set I/C", params.get("Apply I/C", [None] * len(Iranges)))
    # Arrange all the data into the correct format.
    # TODO: Metadata could be handled in a nicer way.
    metadata = {"settings": settings, "params": params}

    ds = process_data(data_lines, Eranges, Iranges, controls, locale)
    if "time" in ds:
        ds["uts"] = ds["time"] + start_time
    else:
        ds["uts"] = [start_time]
    if fulldate:
        del ds.attrs["fulldate"]
    ds.attrs["original_metadata"] = metadata

    return DataTree(ds)



"""
Processing of ULVAC PHI Multipak XPS traces.

The `IGOR .spe import script <https://www.wavemetrics.com/project/phispefileloader>`_ by
jjweimer was pretty helpful for writing this extractor.

Usage
`````
Available since ``yadg-4.0``.

.. autopydantic_model:: dgbowl_schemas.yadg.dataschema_5_1.filetype.Phi_spe

Schema
``````
.. code-block:: yaml

    datatree.DataTree:
      {{ trace_name }}:
        coords:
          E:            !!float               # Binding energies
        data_vars:
          y:            (E)                   # Signal data

Metadata
````````
The following metadata is extracted:

    - ``software_id``: ID of the software used to generate the file.
    - ``version``: Version of the software used to generate the file.
    - ``username``: User name used to generate the file.

Additionally, the processed header data is stored in the metadata under ``file_header``.

Notes on file structure
```````````````````````

These binary files actually contain an ASCII file header, delimited by
`"SOFH\n"` and `"EOFH\n"`.

The binding energies corresponding to the datapoints in the later part
of the file can be found from the `"SpectralRegDef"` entries in this
header. Each of these entries look something like:

.. code-block::

    2 2 F1s 9 161 -0.1250 695.0 675.0 695.0 680.0    0.160000 29.35 AREA

This maps as follows:

.. code-block::

    2           trace_number
    2           trace_number (again?)
    F1s         name
    9           atomic_number
    161         num_datapoints
    -0.1250     step
    695.0       start
    675.0       stop
    695.0       ?
    680.0       ?
    0.160000    dwell_time
    29.35       e_pass
    AREA        description (?)

After the file header, the binary part starts with a short data header
(offsets given from start of data header):

.. code-block::

    0x0000 group                # Data group number.
    0x0004 num_traces           # Number of traces in file
    0x0008 trace_header_size    # Combined lengths of all trace headers.
    0x000c data_header_size     # Length of this data header.

After this follow ``num_traces`` trace headers that are each structured
something like this:

.. code-block::

    0x0000 trace_number          # Number of the trace.
    0x0004 bool_01               # ???
    0x0008 bool_02               # ???
    0x000c trace_number_again    # Number of the trace. Again?
    0x0010 bool_03               # ???
    0x0014 num_datapoints        # Number of datapoints in trace.
    0x0018 bool_04               # ???
    0x001c bool_05               # ???
    0x0020 string_01             # ???
    0x0024 string_02             # ???
    0x0028 string_03             # ???
    0x002c int_02                # ???
    0x0030 string_04             # ???
    0x0034 string_05             # ???
    0x0038 y_unit                # The unit of the datapoints.
    0x003c int_05                # ???
    0x0040 int_06                # ???
    0x0044 int_07                # ???
    0x0048 data_dtype            # Data type for datapoints (f4 / f8).
    0x004c num_data_bytes        # Unsure about this one.
    0x0050 num_datapoints_tot    # This one as well.
    0x0054 int_10                # ???
    0x0058 int_11                # ???
    0x005c end_of_data           # Byte offset of the end-of-data.

After the trace headers follow the datapoints. After the number of
datapoints there is a single 32bit float with the trace's dwelling time
again.

Uncertainties
`````````````
The uncertainties of ``"E"`` are taken as the step-width of
the linearly spaced energy values.

The uncertainties ``"s"`` of ``"y"`` are currently set to a constant
value of ``12.5`` counts per second as all the signals in the files seen so
far only seem to take on values in those steps.

.. admonition:: TODO

    https://github.com/dgbowl/yadg/issues/13

    Determining the uncertainty of the counts per second signal in XPS
    traces from the phispe parser should be done in a better way.

.. codeauthor::
    Nicolas Vetsch

"""

import re
import numpy as np
import xarray as xr
import datatree
from datatree import DataTree
import yadg.dgutils as dgutils

data_header_dtype = np.dtype(
    [
        ("group", "<u4"),
        ("num_traces", "<u4"),
        ("trace_header_size", "<u4"),
        ("data_header_size", "<u4"),
    ]
)

trace_header_dtype = np.dtype(
    [
        ("trace_number", "<u4"),
        ("bool_01", "<u4"),
        ("bool_02", "<u4"),
        ("trace_number_again", "<u4"),
        ("bool_03", "<u4"),
        ("num_datapoints", "<u4"),
        ("bool_04", "<u4"),
        ("bool_05", "<u4"),
        ("string_01", "|S4"),  # pnt?
        ("string_02", "|S4"),
        ("string_03", "|S4"),  # sar?
        ("int_02", "<u4"),
        ("string_04", "|S4"),
        ("string_05", "|S4"),
        ("y_unit", "|S4"),
        ("int_05", "<u4"),
        ("int_06", "<u4"),
        ("int_07", "<u4"),
        ("data_dtype", "|S4"),
        ("num_data_bytes", "<u4"),
        ("num_datapoints_tot", "<u4"),
        ("int_10", "<u4"),
        ("int_11", "<u4"),
        ("end_of_data", "<u4"),
    ]
)


def camel_to_snake(s: str) -> str:
    """Converts CamelCase strings to snake_case.

    From https://stackoverflow.com/a/1176023

    Parameters
    ----------
    s
        The CamelCase input string.

    Returns
    -------
    str
        The snake_case equivalent of s.

    """
    s = re.sub(r"(.)([A-Z][a-z]+)", r"\1_\2", s)
    return re.sub(r"([a-z0-9])([A-Z])", r"\1_\2", s).lower()


def _process_header(spe: list[bytes]) -> dict:
    """Processes the file header at the top of `.spe` files.

    Parameters
    ----------
    spe
        The lines of bytes read from file.

    Returns
    -------
    dict
        The file header parsed into a dictionary. Some entries (keys)
        occur more than once. The corresponding values are joined into
        a list.

    """
    header_lines = spe[spe.index(b"SOFH\n") + 1 : spe.index(b"EOFH\n")]
    header = {}
    for line in header_lines:
        key, value = line.split(b":")
        key, value = camel_to_snake(key.decode().strip()), value.decode().strip()
        if key in header:
            header[key] = [header[key]] + [value]
        else:
            header[key] = value
    return header


def _process_trace_defs(header: dict) -> list[dict]:
    """Processes the trace definition strings given in the file header.

    These strings look something like the following:
    `2 2 F1s 9 161 -0.1250 695.0 675.0 695.0 680.0    0.160000 29.35 AREA`

    Parameters
    ----------
    header
        The file header parsed into a dictionary. The "SpectralRegDef"
        entry contains a list of trace definition strings.

    Returns
    -------
    list[dict]
        A list of trace definition dictionaries describind the kind of
        trace and the binding energy ranges.

    """
    trace_defs = []
    for trace_def in header.get("spectral_reg_def"):
        trace_def = trace_def.split()
        trace_defs.append(
            {
                "trace_number": int(trace_def[0]),
                "name": trace_def[2],
                "atomic_number": trace_def[3],
                "num_datapoints": int(trace_def[4]),
                "step": float(trace_def[5]),
                "start": float(trace_def[6]),
                "stop": float(trace_def[7]),
                "dwell_time": trace_def[10],
                "e_pass": trace_def[11],
                "description": trace_def[12],
            }
        )
    return trace_defs


def _process_traces(spe: list[bytes], trace_defs: list[dict]) -> dict:
    """Processes the spectral traces in the file.

    Parameters
    ----------
    spe
        The lines of bytes read from file.

    trace_defs
        The list of trace definitions parsed from the file header.

    Returns
    -------
    dict
        A dictionary containing the binding energies constructed from
        the trace definitions and the corrresponding XPS traces.

    """
    data = b"".join(spe[spe.index(b"EOFH\n") + 1 :])
    data_header = dgutils.read_value(data, 0x0000, data_header_dtype)
    assert data_header["num_traces"] == len(trace_defs)
    # All trace headers I have seen are 192 (0xc0) bytes long.
    assert data_header["trace_header_size"] / trace_header_dtype.itemsize == len(
        trace_defs
    )
    assert data_header["data_header_size"] == data_header_dtype.itemsize
    trace_headers = np.frombuffer(
        data,
        offset=0x0010,
        dtype=trace_header_dtype,
        count=len(trace_defs),
    )
    traces = {}
    for trace_header, trace_def in zip(trace_headers, trace_defs):
        assert trace_header["trace_number"] == trace_def["trace_number"]
        assert trace_header["num_datapoints"] == trace_def["num_datapoints"]
        # Contruct the binding energies from trace_def.
        energies, dE = np.linspace(
            trace_def["start"],
            trace_def["stop"],
            trace_def["num_datapoints"],
            endpoint=True,
            retstep=True,
        )
        # Construct data from trace_header
        data_dtype = np.dtype(f'{trace_header["data_dtype"].decode()}')
        data_offset = trace_header["end_of_data"] - trace_header["num_data_bytes"]
        datapoints = np.frombuffer(
            data,
            offset=data_offset,
            dtype=data_dtype,
            count=trace_header["num_datapoints"],
        )
        dwell_time = dgutils.read_value(data, trace_header["end_of_data"], "<f4")
        np.testing.assert_almost_equal(dwell_time, float(trace_def["dwell_time"]))
        # TODO: Figure out the correct error. This signal count should
        # somehow be a Poisson distribution, i.e. the error should be
        # something like s ~ sqrt(n). The counts per second only seem to
        # be taking values in steps of 12.5cps, so taking this as error.
        traces[str(trace_def["trace_number"])] = {
            "name": trace_def["name"],
            "atomic_number": trace_def["atomic_number"],
            "dwell_time": trace_def["dwell_time"],
            "e_pass": trace_def["e_pass"],
            "description": trace_def["description"],
            "yvals": datapoints,
            "ydevs": np.ones(len(datapoints)) * 12.5,
            "yunit": "counts / s",
            "Evals": energies,
            "Edevs": np.ones(len(energies)) * abs(dE),
            "Eunit": "eV",
        }
    return traces


def extract(
    *,
    fn: str,
    **kwargs: dict,
) -> DataTree:
    with open(fn, "rb") as spe_file:
        spe = spe_file.readlines()
    header = _process_header(spe)
    software_id, version = header.get("software_version").split()
    meta = {
        "software_id": software_id,
        "version": version,
        "username": header.get("operator"),
        "file_header": header,
    }
    trace_defs = _process_trace_defs(header)
    traces = _process_traces(spe, trace_defs)
    vals = {}
    for v in traces.values():
        fvals = xr.Dataset(
            data_vars={
                "y": (
                    ["E"],
                    v["yvals"],
                    {"units": v["yunit"], "ancillary_variables": "y_std_err"},
                ),
                "y_std_err": (
                    ["E"],
                    v["ydevs"],
                    {"units": v["yunit"], "standard_name": "y standard_error"},
                ),
                "E_std_err": (
                    ["E"],
                    v["Edevs"],
                    {"units": v["Eunit"], "standard_name": "E standard_error"},
                ),
            },
            coords={
                "E": (
                    ["E"],
                    v["Evals"],
                    {"units": v["Eunit"], "ancillary_variables": "E_std_err"},
                ),
            },
        )
        vals[v["name"]] = fvals

    dt = datatree.DataTree.from_dict(vals)
    dt.attrs["original_metadata"] = meta
    return dt



import pytest
import os
import datatree
import yadg.core
from dgbowl_schemas.yadg import to_dataschema, DataSchema_4_0, DataSchema_4_1
from pydantic import ValidationError
from pydantic.v1 import ValidationError as ValidationError_v1
from .utils import datagram_from_file

ts0 = {
    "metadata": {"provenance": "manual", "schema_version": "4.0.1"},
    "steps": [{"parser": "dummy", "import": {"folders": ["."], "suffix": "wrong"}}],
}
ts1 = {
    "metadata": {"provenance": "manual", "schema_version": "4.0.1"},
    "steps": [{"parser": "dummy", "import": {"files": ["ts1.dummy.json"]}}],
}
ts2 = {
    "metadata": {"provenance": "manual", "schema_version": "4.0.1"},
    "steps": [{"parser": "dummy", "import": {"folders": ["."], "contains": ".dummy."}}],
}
ts3 = {
    "metadata": {"provenance": "manual", "schema_version": "4.0.1"},
    "steps": [
        {
            "parser": "dummy",
            "import": {"files": ["ts0.dummy.json", "ts1.dummy.json"]},
        }
    ],
}
ts4 = {
    "metadata": {"provenance": "manual", "schema_version": "4.0.1"},
    "steps": [
        {
            "parser": "dummy",
            "import": {"folders": ["."], "prefix": "ts", "contains": "1"},
        }
    ],
}
fts0 = {
    "metadata": {"provenance": "manual", "schema_version": "4.0.1"},
    "steps": [
        {
            "parse": "dummy",
            "import": {"files": ["dummy_schema_1.json", "dummy_schema_2.json"]},
        }
    ],
}
fts1 = {
    "metadata": {"provenance": "manual", "schema_version": "4.0.1"},
    "steps": [
        {
            "parser": "dumm",
            "import": {"files": ["dummy_schema_1.json", "dummy_schema_2.json"]},
        }
    ],
}
fts2 = {
    "metadata": {"provenance": "manual", "schema_version": "4.0.1"},
    "steps": [
        {
            "parser": "dummy",
            "import": {
                "files": ["dummy_schema_1.json", "dummy_schema_2.json"],
                "folders": ["."],
            },
        }
    ],
}
fts3 = {
    "metadata": {"provenance": "manual", "schema_version": "4.0.1"},
    "steps": [{"parser": "dummy", "import": {}}],
}
fts4 = {
    "metadata": {"provenance": "manual", "schema_version": "4.0.1"},
    "steps": [
        {
            "parser": "dummy",
            "import": {"files": ["dummy_schema_1.json", "dummy_schema_2.json"]},
            "key": "value",
        }
    ],
}

ts5 = {
    "metadata": {"provenance": {"type": "manual"}, "version": "4.1"},
    "steps": [
        {
            "parser": "dummy",
            "input": {
                "folders": ["."],
                "suffix": "json",
                "contains": "schema",
                "exclude": "3.1.0",
            },
            "parameters": {},
        }
    ],
}

fts5 = {
    "metadata": {"provenance": "wrong", "version": "4.1"},
    "steps": [
        {
            "parser": "dummy",
            "input": {"files": ["dummy_schema_1.json", "dummy_schema_2.json"]},
        }
    ],
}
fts6 = {
    "metadata": {"provenance": {"type": "manual"}, "version": "4.0"},
    "steps": [
        {
            "parser": "dummy",
            "input": {"files": ["dummy_schema_1.json", "dummy_schema_2.json"]},
        }
    ],
}
fts7 = {
    "metadata": {"provenance": {"type": "manual"}, "version": "4.1"},
    "steps": [
        {
            "parser": "dummy",
            "import": {"files": ["dummy_schema_1.json", "dummy_schema_2.json"]},
        }
    ],
}
fts8 = {
    "metadata": {"provenance": {"type": "manual"}, "version": "4.1"},
    "steps": [
        {
            "input": {"files": ["dummy_schema_1.json", "dummy_schema_2.json"]},
            "parameters": {"k": "v"},
        }
    ],
}


@pytest.mark.parametrize(
    "inp_dict, l_dg, l_res",
    [
        (ts0, 1, 0),
        (ts1, 1, 1),
        (ts2, 1, 2),
        (ts3, 1, 2),
        (ts4, 1, 1),
        (ts5, 1, 0),
    ],
)
def test_datagram_from_schema_dict(inp_dict, l_dg, l_res, datadir):
    os.chdir(datadir)
    ds = to_dataschema(**inp_dict)
    ret = yadg.core.process_schema(ds)
    print(f"{ret=}")
    assert len(ret.children) == l_dg, "wrong number of steps"
    if l_dg > 0:
        assert len(ret["0"].get("uts", [])) == l_res, "wrong number of timesteps"
    ret.to_netcdf("test.nc", engine="h5netcdf")
    ref = datatree.open_datatree("test.nc", engine="h5netcdf")
    print(ref)
    assert ref == ret


@pytest.mark.parametrize(
    "inp_fn, ts",
    [
        (
            "ts0.dummy.json",
            {"nsteps": 1, "step": "0", "item": 0, "kwargs": {}},
        ),
        (
            "ts1.dummy.json",
            {"nsteps": 2, "step": "1", "item": 0, "kwargs": {"k": "v"}},
        ),
        (
            "ts2.json",
            {"nsteps": 2, "step": "1", "item": 0, "kwargs": {"k": "v"}},
        ),
    ],
)
def test_datagram_from_schema_file(inp_fn, ts, datadir):
    os.chdir(datadir)
    ret = datagram_from_file(inp_fn)
    assert len(ret.children) == ts["nsteps"], (
        "wrong number of steps: " f"got: {len(ret.children)}, expected: {ts['nsteps']}"
    )
    for k, v in ts["kwargs"].items():
        assert ret[ts["step"]][k][ts["item"]] == v, "kwargs not passed correctly"
    ret.to_netcdf("test.nc", engine="h5netcdf")
    ref = datatree.open_datatree("test.nc", engine="h5netcdf")
    print(ref)
    assert ref == ret


@pytest.mark.parametrize(
    "inp_dict, expr",
    [
        (fts0, r"Discriminator 'parser' is missing in value"),
        (fts1, r"No match for discriminator 'parser' and value 'dumm'"),
        (fts2, r"Both 'files' and 'folders'"),
        (fts3, r"Neither 'files' nor 'folders'"),
        (fts4, r"steps -> 0 -> Dummy -> key\n  extra fields not permitted"),
    ],
)
def test_schema_validator_4_0(inp_dict, expr, datadir):
    os.chdir(datadir)
    with pytest.raises((ValidationError, ValidationError_v1), match=expr):
        assert DataSchema_4_0(**inp_dict)


@pytest.mark.parametrize(
    "inp_dict, expr",
    [
        (fts5, r"metadata -> provenance\n  value is not a valid dict"),
        (fts6, r"metadata -> version\n  unexpected value"),
        (fts7, r"steps -> 0 -> Dummy -> input\n  field required"),
        (fts8, r"Discriminator 'parser' is missing in value"),
    ],
)
def test_schema_validator_4_1(inp_dict, expr, datadir):
    os.chdir(datadir)
    with pytest.raises((ValidationError, ValidationError_v1), match=expr):
        assert DataSchema_4_1(**inp_dict)



import pytest
import os
import pickle
from yadg.extractors.empalc.xlsx import extract
from .utils import compare_datatrees, datagram_from_file


@pytest.mark.parametrize(
    "infile",
    [
        "Cu-25p_v2.xlsx",
        "samplename_newlines.xlsx",
        "2022-09-12-15-37-07+0200_spCuDurapore05_old_injections_LC-data.xlsx",
    ],
)
def test_empalc_xlsx(infile, datadir):
    os.chdir(datadir)
    ret = extract(fn=infile)
    outfile = f"{infile}.pkl"
    with open(outfile, "rb") as inp:
        ref = pickle.load(inp)
    print(f"{ret=}")
    with open(outfile, "wb") as out:
        pickle.dump(ret, out, 5)
    compare_datatrees(ret, ref, thislevel=True)


def test_empalc_lock_stock_dataschema(datadir):
    os.chdir(datadir)
    ret = datagram_from_file("lock_stock_dataschema.yml")
    print(f"{ret=}")
    for k in {"height", "concentration", "retention time", "area"}:
        assert ret["LC"][k].shape == (7, 2)



"""
Handles processing of Panalytical XRDML files.

Usage
`````
Available since ``yadg-4.2``.

.. autopydantic_model:: dgbowl_schemas.yadg.dataschema_5_1.filetype.Panalytical_xrdml

Schema
``````
.. code-block:: yaml

    datatree.DataTree:
      coords:
        uts:            !!float               # Unix timestamp
        angle:          !!float               # 2θ angle
      data_vars:
        intensity:      (uts, angle)          # Measured intensity

Metadata
````````
The following metadata is extracted:

  - ``sample``: Metadata information about the sample.
  - ``wavelength``: Measurement wavelength.
  - ``comment``: A free-form description of the experiment.
  - ``incident_beam_path``
  - ``diffracted_beam_path``
  - ``counting_time``

.. note::

    The returned metadata contain only a subset of the available metadata in the XML
    file. If something important is missing, please contact us!

Notes on file structure
```````````````````````
These are xml-formatted files, which we here parse using the :mod:`xml.etree`
library into a Python :class:`dict`.

The ``angle`` returned from this parser is based on a linear interpolation of the start
and end point of the scan, and is the :math:`2\\theta`. The values of :math:`\\omega`
are discarded.

Uncertainties
`````````````
The uncertainties of in ``"angle"`` are taken as the step-width of the linearly spaced
:math:`2\\theta` values.

The uncertainties of of ``"intensity"`` are currently set to a constant
value of 1.0 count as all the supported files seem to produce integer values.

.. codeauthor::
    Nicolas Vetsch,
    Peter Kraus

"""

from collections import defaultdict
from typing import Union
from xml.etree import ElementTree
import numpy as np
from datatree import DataTree
import xarray as xr
from uncertainties.core import str_to_number_with_uncert as tuple_fromstr

from yadg.extractors.panalytical.common import panalytical_comment
from yadg.dgutils import dateutils


def etree_to_dict(e: ElementTree.Element) -> dict:
    """Recursively converts an ElementTree.Element into a dictionary.

    Element attributes are stored into `"@"`-prefixed attribute keys.
    Element text is stored into `"#text"` for all nodes.

    From https://stackoverflow.com/a/10076823.
    """
    d = {e.tag: {} if e.attrib else None}
    children = list(e)
    if children:
        dd = defaultdict(list)
        for dc in map(etree_to_dict, children):
            for k, v in dc.items():
                dd[k].append(v)
        d = {e.tag: {k: v[0] if len(v) == 1 else v for k, v in dd.items()}}
    if e.attrib:
        d[e.tag].update(("@" + k, v) for k, v in e.attrib.items())
    if e.text:
        text = e.text.strip()
        if children or e.attrib:
            if text:
                d[e.tag]["#text"] = text
        else:
            d[e.tag] = text
    return d


def _process_values(d: Union[dict, str]) -> Union[dict, str]:
    """
    Recursively parses dicts in the following format:

    .. code::

        {"key": {"#text": ..., "@unit": ...}, ...}

    into a single string:

    .. code::

        {"key": f"{#text} {@unit}", ...}
    """
    # TODO
    # If not "#text" or @tribute just snake_case and recurse.
    if isinstance(d, dict):
        if "@unit" in d and "#text" in d:
            return f"{d['#text']} {d['@unit']}"
        elif "@version" in d and "#text" in d:
            return f"{d['#text']} {d['@version']}"
        else:
            for k, v in d.items():
                d[k] = _process_values(v)
    return d


def _process_scan(scan: dict) -> dict:
    """
    Parses the scan section of the file. Creates the explicit positions based
    on the number of measured intensities and the start & end position.
    """
    header = scan.pop("header")
    dpts = scan.pop("dataPoints")
    counting_time = _process_values(dpts.pop("commonCountingTime"))
    ivals, idevs = list(
        zip(*[tuple_fromstr(c) for c in dpts["intensities"].pop("#text").split()])
    )
    iunit = dpts["intensities"].pop("@unit")
    timestamp = header.pop("startTimeStamp")

    dp = {
        "intensity": {"vals": ivals, "devs": idevs, "unit": iunit},
        "timestamp": timestamp,
        "counting_time": counting_time,
    }

    positions = _process_values(dpts.pop("positions"))
    for v in positions:
        pos = np.linspace(
            float(v["startPosition"]), float(v["endPosition"]), num=len(ivals)
        )
        adiff = np.abs(np.diff(pos)) * 0.5
        adiff = np.append(adiff, adiff[-1])
        dp[v["@axis"]] = {
            "vals": pos,
            "devs": adiff,
            "unit": v["@unit"],
        }
    return dp


def _process_comment(comment: dict) -> dict:
    entry = comment.pop("entry")
    ret = {}
    for line in entry:
        ret.update(panalytical_comment(line))
    return ret


def _process_measurement(measurement: dict, timezone: str):
    """
    A function that processes each section of the XRD XML file.
    """
    # Comment.
    comment = measurement["comment"].pop("entry")
    for line in comment:
        if "PHD Lower Level" in line and "PHD Upper Level" in line:
            __, values = list(zip(*[s.split(" = ") for s in line.split(", ")]))
    keys = ["phd_lower_level", "phd_upper_level"]
    measurement["comment"] = dict(zip(keys, values))
    # Wavelength.
    wavelength = _process_values(measurement.pop("usedWavelength"))
    measurement["wavelength"] = wavelength
    # Incident beam path.
    incident_beam_path = _process_values(measurement.pop("incidentBeamPath"))
    measurement["incident_beam_path"] = incident_beam_path
    # Diffracted beam path.
    diffracted_beam_path = _process_values(measurement.pop("diffractedBeamPath"))
    measurement["diffracted_beam_path"] = diffracted_beam_path
    scan = _process_scan(measurement.pop("scan"))
    trace = {"angle": scan.pop("2Theta"), "intensity": scan.pop("intensity")}
    meta = measurement
    meta["counting_time"] = scan.pop("counting_time")
    trace["uts"] = dateutils.str_to_uts(
        timestamp=scan.pop("timestamp"), timezone=timezone
    )
    return trace, meta


def extract(
    *,
    fn: str,
    timezone: str,
    **kwargs: dict,
) -> DataTree:
    it = ElementTree.iterparse(fn)
    # Removing xmlns prefixes from all tags.
    # From https://stackoverflow.com/a/25920989.
    for __, e in it:
        __, xmlns_present, postfix = e.tag.partition("}")
        if xmlns_present:
            e.tag = postfix  # Strip away all xmlns prefixes.
    root = it.root
    xrd = etree_to_dict(root)
    # Start processing the xml contents.
    measurements = xrd["xrdMeasurements"]
    assert measurements["@status"] == "Completed", "Incomplete measurement."
    comment = _process_comment(measurements["comment"])
    # Renaming some entries because I want to.
    sample = measurements["sample"]
    sample["prepared_by"] = sample.pop("preparedBy")
    sample["type"] = sample.pop("@type")
    # Process measurement data.
    data, meta = _process_measurement(measurements["xrdMeasurement"], timezone)
    data["fn"] = fn
    # Shove unused data into meta
    meta["sample"] = sample
    meta["comment"] = comment
    meta["fulldate"] = True
    # Build Datasets
    vals = xr.Dataset(
        data_vars={
            "intensity": (
                ["uts", "angle"],
                np.reshape(data["intensity"]["vals"], (1, -1)),
                {
                    "units": data["intensity"]["unit"],
                    "ancillary_variables": "intensity_std_err",
                },
            ),
            "intensity_std_err": (
                ["uts", "angle"],
                np.reshape(data["intensity"]["devs"], (1, -1)),
                {
                    "units": data["intensity"]["unit"],
                    "standard_name": "intensity standard_error",
                },
            ),
            "angle_std_err": (
                ["uts", "angle"],
                np.reshape(data["angle"]["devs"], (1, -1)),
                {
                    "units": data["angle"]["unit"],
                    "standard_name": "angle standard_error",
                },
            ),
        },
        coords={
            "uts": (["uts"], [data["uts"]]),
            "angle": (
                ["angle"],
                data["angle"]["vals"],
                {
                    "units": data["angle"]["unit"],
                    "ancillary_variables": "angle_std_err",
                },
            ),
        },
        attrs=dict(original_metadata=meta),
    )
    return DataTree(vals)



"""
Handles processing of csv exports of Panalytical XRD files.

Usage
`````
Available since ``yadg-4.2``.

.. autopydantic_model:: dgbowl_schemas.yadg.dataschema_5_1.filetype.Panalytical_csv

Schema
``````
.. code-block:: yaml

    datatree.DataTree:
      coords:
        uts:            !!float               # Unix timestamp
        angle:          !!float               # 2θ angle
      data_vars:
        intensity:      (uts, angle)          # Measured intensity

Metadata
````````
With the exception of the ``comment``, the metadata present in the csv file is extracted
from the file header without post-processing.

Notes on file structure
```````````````````````
These files are split into a ``[Measurement conditions]`` and a ``[Scan points]``
section. The former stores the metadata and the latter all the datapoints.

Uncertainties
`````````````
The uncertainties of ``"angle"`` are taken from the number of significant figures.

The uncertainties of ``"intensity"`` are taken from the number of significant figures.

.. codeauthor::
    Nicolas Vetsch,
    Peter Kraus
"""

from uncertainties.core import str_to_number_with_uncert as tuple_fromstr
from datatree import DataTree
import xarray as xr
import numpy as np

from yadg.dgutils import dateutils
from yadg.extractors.panalytical.common import panalytical_comment, snake_case


def _process_comments(comments: list[str]) -> dict:
    ret = {}
    for line in comments:
        ret.update(panalytical_comment(line))
    return ret


def _process_header(header: str) -> dict:
    """
    Processes the header section, staring with the ``[Measurement conditions]`` line.

    Parameters
    ----------
    header
        The header portion as a string.

    Returns
    -------
    header: dict
        A dictionary containing the processed metadata.

    """
    header_lines = header.split("\n")[1:-1]
    header = dict([line.split(",", 1) for line in header_lines])
    # Process comment entries.
    comments = []
    for key in list(header.keys()):
        if key.startswith("Comment"):
            comments.append(header.pop(key).strip('"'))
    comments = _process_comments(comments)
    # Renaming the keys.
    for key in list(header.keys()):
        header[snake_case(key)] = header.pop(key)
    header.update(comments)
    return header


def _process_data(data: str) -> tuple[list, list]:
    """
    Processes the data section, starting with the ``[Scan points]`` line.

    Parameters
    ----------
    data
        The data portion as a string.

    Returns
    -------
    avals, adevs, ivals, idevs
        The values and uncertainties in angle and intensity.

    """
    data_lines = data.split("\n")[1:-1]
    columns = data_lines[0].split(",")
    assert columns == ["Angle", "Intensity"], "Unexpected columns."
    datapoints = [line.split(",") for line in data_lines[1:]]
    angle, intensity = [list(d) for d in zip(*datapoints)]
    avals, adevs = list(zip(*[tuple_fromstr(a) for a in angle]))
    ivals, idevs = list(zip(*[tuple_fromstr(i) for i in intensity]))
    return avals, adevs, ivals, idevs


def extract(
    *,
    fn: str,
    encoding: str,
    timezone: str,
    **kwargs: dict,
) -> DataTree:
    with open(fn, "r", encoding=encoding) as csv_file:
        csv = csv_file.read()
    # Split file into its sections.
    __, header, data = csv.split("[")
    assert header.startswith("Measurement conditions"), "Unexpected section."
    assert data.startswith("Scan points"), "Unexpected section."
    header = _process_header(header)
    # Process the data trace.
    angle, _, insty, _ = _process_data(data)
    adiff = np.abs(np.diff(angle)) * 0.5
    adiff = np.append(adiff, adiff[-1])
    idevs = np.ones(len(insty))
    # Process the metadata.
    uts = dateutils.str_to_uts(
        timestamp=header["file_date_and_time"],
        format="%d/%B/%Y %H:%M",
        timezone=timezone,
    )
    header["fulldate"] = True
    # Build Datasets
    vals = xr.Dataset(
        data_vars={
            "intensity": (
                ["uts", "angle"],
                np.reshape(insty, (1, -1)),
                {"units": "counts", "ancillary_variables": "intensity_std_err"},
            ),
            "intensity_std_err": (
                ["uts", "angle"],
                np.reshape(idevs, (1, -1)),
                {"units": "counts", "standard_name": "intensity standard_error"},
            ),
            "angle_std_err": (
                ["uts", "angle"],
                np.reshape(adiff, (1, -1)),
                {"units": "deg", "standard_name": "angle standard_error"},
            ),
        },
        coords={
            "uts": (["uts"], [uts]),
            "angle": (
                ["angle"],
                list(angle),
                {"units": "deg", "ancillary_variables": "angle_std_err"},
            ),
        },
        attrs=dict(original_metadata=header),
    )
    return DataTree(vals)



"""
This module parses Touchstone files conforming to the Touchstone File Specification,
`revision 1.1 <https://ibis.org/connector/touchstone_spec11.pdf>`_.

.. note::

    Currently only 1- and 2-port (i.e. ``.s1p`` and ``.s2p``) files are supported.

Usage
`````
Available since ``yadg-5.1``.

.. autopydantic_model:: dgbowl_schemas.yadg.dataschema_5_1.filetype.Touchstone_snp

Schema
``````
.. code-block:: yaml

    datatree.DataTree:
      {{ parameter_name }}                  # S11, S21, etc. for scattering parameters
        coords:
          uts:            !!float           # Unix timestamp, optional
          frequency:      !!float           #
        data_vars:
          real:           (uts, frequency)  # Real part of the response
          imag:           (uts, frequency)  # Imagunary part of the response
          magnitude:      (uts, frequency)  # Magnitude of the response
          phase_angle:    (uts, frequency)  # Phase angle of the response

Metadata
````````
Metadata about the device are inferred from the structure of the comments lines, see
below. In most circumstances, the following information can be provided:

- ``Ref R``: the reference resistance (in Ohms)
- ``Model``: model number of the network analyseer
- ``Serial``: serial number of the device
- ``Software Version``: software version with which the file was created

.. note::

    The timestamp (``uts``) is also parsed from the comments in the file. In case you
    have Touchstone files with a well-defined header that is not supported by yadg,
    please open an issue.

Uncertainties
`````````````
Uncertainties in all variables are determined from the precision of the string-to-float
conversion.

Notes on file structure
```````````````````````
The Touchstone ``.sNp`` files are composed of four sections:

- options line, denoted by ``#``
- any number of comments lines, denoted by ``!``
- data lines
- a noise parameters section, delimited by ``! NOISE PARAMETERS``

Currently, only the first three sections are parsed.

.. codeauthor::
    Peter Kraus

"""

import logging
from datatree import DataTree
from xarray import Dataset, DataArray
from babel.numbers import parse_decimal

from yadg import dgutils

logger = logging.getLogger(__name__)


def process_filename(filename: str) -> dict:
    if filename.lower().endswith(".s1p"):
        params = ["_11"]
    elif filename.lower().endswith(".s2p"):
        params = ["_11", "_21", "_12", "_22"]
    else:
        logger.warning("Could not determine number of ports, assuming 1 port.")
        params = ["_11"]
    return {"params": params}


def process_options(line: str) -> dict:
    _, freq_unit, param, format, _, Rref = line.split()
    if format == "DB" or format == "dB":
        columns = ["magnitude", "phase_angle"]
        mag_unit = "dB"
    elif format == "MA":
        columns = ["magnitude", "phase_angle"]
        mag_unit = None
    elif format == "RI":
        columns = ["real", "imag"]
        mag_unit = None
    else:
        raise RuntimeError(f"Unsupported format: {format!r}")

    return {
        "freq_unit": freq_unit.replace("HZ", "Hz"),
        "param": param,
        "columns": columns,
        "mag_unit": mag_unit,
        "Rref": float(Rref),
    }


def attr_in_lines(attr: str, lines: list[str], sep: str = ":") -> dict:
    for line in lines:
        if line.startswith(attr):
            return {attr: line[line.index(sep) + 1 :].strip()}
    return {}


def process_comments(lines: list[str], tz: str) -> dict:
    attrs = {}
    uts = None
    comstr = "\n".join(lines)

    # No header - pass
    if len(lines) == 0:
        pass
    # PicoVNA 108 files
    elif "Ref Plane:" in comstr and len(lines) == 3:
        fmt = "%m/%d/%Y  %I:%M:%S %p"
        uts = dgutils.str_to_uts(
            timestamp=lines[0], timezone=tz, format=fmt, strict=False
        )
        attrs["Ref Plane"] = lines[1].split(":")[1].strip()
        attrs["Model"] = "PicoVNA 108"
    # Agilent N523* and E50* export
    elif "Agilent Technologies" in lines[0] or "Keysight Technologies" in lines[0]:
        _, model, serial, version = lines[0].split(",")
        attrs["Model"] = model
        attrs["Serial"] = serial
        attrs["Software Version"] = version
        attrs.update(attr_in_lines("Correction", lines, ":"))
        datestr = attr_in_lines("Date", lines, ":")["Date"]
        for fmt in ["%A, %B %d, %Y %H:%M:%S", "%a %b %d %H:%M:%S %Y"]:
            uts = dgutils.str_to_uts(
                timestamp=datestr, timezone=tz, format=fmt, strict=False
            )
            if uts is not None:
                break
    # Agilent N991* export
    elif "Agilent N99" in lines[0] or "Keysight N99" in lines[0]:
        attrs.update(attr_in_lines("Model", lines, ":"))
        attrs.update(attr_in_lines("Serial", lines, ":"))
        datestr = attr_in_lines("Date", lines, ":")["Date"]
        fmt = "%A, %d %B %Y %H:%M:%S"
        uts = dgutils.str_to_uts(
            timestamp=datestr, timezone=tz, format=fmt, strict=False
        )
    # Rohde & Schwarz ZVA/ZVB export
    elif "Rohde & Schwarz" in lines[0]:
        model, version, serial = lines[0].split(" - ")
        attrs["Model"] = model
        attrs["Serial"] = serial
        attrs["Software Version"] = version
        datestr = attr_in_lines("Date", lines, ":")["Date"]
        fmt = "%Y-%m-%d %H:%M:%S"
        uts = dgutils.str_to_uts(
            timestamp=datestr, timezone=tz, format=fmt, strict=False
        )
    # Unknown header - pass
    else:
        pass
    return uts, attrs


def data_to_dataset(key: str, data: dict) -> Dataset:
    coords = {"frequency": data["frequency"]}
    data_vars = {
        "frequency_std_err": DataArray(
            data=data["frequency_std_err"], dims=["frequency"]
        )
    }
    for k, v in data[key].items():
        data_vars[k] = DataArray(data=v, dims=coords.keys())
    return Dataset(data_vars=data_vars, coords=coords)


def extract(
    *,
    fn: str,
    encoding: str,
    timezone: str,
    locale: str,
    **kwargs: dict,
) -> DataTree:
    with open(fn, encoding=encoding) as inf:
        lines = inf.readlines()
    metadata = process_filename(fn)

    # Find options line
    for li in lines:
        if li.startswith("#"):
            metadata.update(process_options(li))
            # Only one options line is allowed - the rest is ignored, hence break
            break

    # Prepare data structures
    metadata["params"] = [p.replace("_", metadata["param"]) for p in metadata["params"]]
    cols = ["frequency"]
    data = {"frequency": [], "frequency_std_err": []}
    for k in metadata["params"]:
        data[k] = {}
        for m in metadata["columns"]:
            data[k][m] = []
            data[k][f"{m}_std_err"] = []
            cols.append((k, m))
    comments = []

    # Parse data lines
    for li in lines:
        if li.startswith("#"):
            continue
        elif li.startswith("!"):
            # Watch out for "! NOISE PARAMETERS"
            if "NOISE PARAMETERS" in li:
                break
            else:
                comments.append(li[1:].strip())
        else:
            # Trim comments from data lines
            if "!" in li:
                parts = li[: li.index("!")].split()
            else:
                parts = li.split()

            # For .s3p and above, we'd have to append multiple lines together here
            for k, v in zip(cols, parts):
                dec = parse_decimal(v, locale=locale)
                exp = dec.as_tuple().exponent
                val = float(dec)
                dev = 10**exp
                if k == "frequency":
                    data[k].append(val)
                    data[f"{k}_std_err"].append(dev)
                elif k[1] == "magnitude":
                    data[k[0]][k[1]].append(val)
                    data[k[0]][f"{k[1]}_std_err"].append(dev)
                else:
                    data[k[0]][k[1]].append(val)
                    data[k[0]][f"{k[1]}_std_err"].append(dev)

    uts, attrs = process_comments(comments, timezone)
    attrs["Ref R"] = f"{metadata['Rref']} Ohm"

    dtdict = {"/": Dataset(attrs=dict(original_metadata=attrs))}
    for k in metadata["params"]:
        ds = data_to_dataset(key=k, data=data)
        for var in ds.variables:
            if f"{var}_std_err" in ds.variables:
                ds[var].attrs["ancillary_variables"] = f"{var}_std_err"
            elif var.endswith("_std_err"):
                end = var.index("_std_err")
                if var[:end] in ds.variables:
                    ds[var].attrs["standard_name"] = f"{var[:end]} standard_error"
            if "angle" in var:
                ds[var].attrs["units"] = "degree"
            elif "frequency" in var:
                ds[var].attrs["units"] = metadata["freq_unit"]
            elif "magnitude" in var and metadata["mag_unit"] is not None:
                ds[var].attrs["units"] = metadata["mag_unit"]
        if uts is not None:
            ds = ds.expand_dims(dim=dict(uts=[uts]))
        else:
            ds.attrs["fulldate"] = False
        dtdict[f"/{k}"] = ds
    dt = DataTree.from_dict(dtdict)
    return dt



"""
For processing Inficon Fusion json data.  Contains both the data from the raw
chromatogram and the post-processed results.

Usage
`````
Available since ``yadg-4.0``.

.. autopydantic_model:: dgbowl_schemas.yadg.dataschema_5_1.filetype.Fusion_json

Schema
``````
.. code-block:: yaml

    datatree.DataTree:
      coords:
        uts:              !!float
        species:          !!str
      data_vars:
        height:           (uts, species)        # Peak height at maximum
        area:             (uts, species)        # Integrated peak area
        concentration:    (uts, species)        # Calibrated peak area
        xout:             (uts, species)        # Mole fraction (normalized conc.)
        retention time:   (uts, species)        # Peak retention time
      {{ detector_name }}:
        coords:
          uts:            !!float               # Unix timestamp
          elution_time:   !!float               # Elution time
        data_vars:
          signal:         (uts, elution_time)   # Signal data
          valve:          (uts)                 # Valve position

Metadata
````````
No metadata is currently extracted.

.. codeauthor::
    Peter Kraus

"""

import json
import logging
from datatree import DataTree
from xarray import Dataset
import xarray as xr
import numpy as np

from yadg import dgutils


logger = logging.getLogger(__name__)


def chromdata(jsdata: dict, uts: float) -> Dataset:
    metadata = {
        "method": jsdata.get("methodName", "n/a"),
        "version": jsdata.get("softwareVersion", {}).get("version", None),
        "datafile": jsdata.get("sequence", {}).get("location", None),
    }

    sampleid = jsdata.get("annotations", {}).get("name", None)
    if sampleid is not None:
        metadata["sampleid"] = sampleid

    units = {
        "height": None,
        "area": None,
        "concentration": "%",
        "xout": "%",
        "retention time": "s",
    }

    raw = {
        "height": {},
        "area": {},
        "concentration": {},
        "xout": {},
        "retention time": {},
    }

    species = set()

    # sort detector keys to ensure alphabetic order for ID matching
    for detname in sorted(jsdata["detectors"].keys()):
        detdict = jsdata["detectors"][detname]
        if "analysis" in detdict:
            for peak in detdict["analysis"]["peaks"]:
                if "label" not in peak:
                    continue
                else:
                    species.add(peak["label"])
                if "height" in peak:
                    raw["height"][peak["label"]] = (float(peak["height"]), 1.0)
                if "area" in peak:
                    raw["area"][peak["label"]] = (float(peak["area"]), 0.01)
                if "concentration" in peak:
                    raw["concentration"][peak["label"]] = (
                        float(peak["concentration"]),
                        float(peak["concentration"]) * 1e-3,
                    )
                if "normalizedConcentration" in peak:
                    raw["xout"][peak["label"]] = (
                        float(peak["normalizedConcentration"]),
                        float(peak["normalizedConcentration"]) * 1e-3,
                    )
                if "top" in peak:
                    raw["retention time"][peak["label"]] = (float(peak["top"]), 0.01)

    valve = jsdata.get("annotations", {}).get("valcoPosition", None)
    if valve is not None:
        raw["valve"] = valve

    species = sorted(species)
    data_vars = {}
    for k, v in units.items():
        vals, devs = zip(*[raw[k].get(s, (np.nan, np.nan)) for s in species])
        data_vars[k] = (
            ["uts", "species"],
            [vals],
            {"ancillary_variables": f"{k}_std_err"},
        )
        data_vars[f"{k}_std_err"] = (
            ["uts", "species"],
            [devs],
            {"standard_name": f"{k} stdandard_error"},
        )
        if v is not None:
            data_vars[k][2]["units"] = v
            data_vars[f"{k}_std_err"][2]["units"] = v

    ds = xr.Dataset(
        data_vars=data_vars,
        coords={"species": (["species"], species), "uts": (["uts"], [uts])},
        attrs=dict(original_metadata=metadata),
    )
    return ds


def chromtrace(jsdata: dict, uts: float) -> DataTree:
    metadata = {
        "method": jsdata.get("methodName", "n/a"),
        "sampleid": jsdata.get("annotations", {}).get("name", None),
        "version": jsdata.get("softwareVersion", {}).get("version", None),
        "datafile": jsdata.get("sequence", {}).get("location", None),
    }

    # sort detector keys to ensure alphabetic order for ID matching
    traces = sorted(jsdata["detectors"].keys())
    vals = {}
    for detname in traces:
        detdict = jsdata["detectors"][detname]
        fvals = xr.Dataset(
            data_vars={
                "signal": (
                    ["uts", "elution_time"],
                    [detdict["values"]],
                    {"ancillary_variables": "signal_std_err"},
                ),
                "signal_std_err": (
                    ["uts", "elution_time"],
                    [np.ones(detdict["nValuesExpected"])],
                    {"standard_name": "signal standard_error"},
                ),
                "elution_time_std_err": (
                    ["elution_time"],
                    np.ones(detdict["nValuesExpected"]) / detdict["nValuesPerSecond"],
                    {"units": "s", "standard_name": "elution_time standard_error"},
                ),
            },
            coords={
                "elution_time": (
                    ["elution_time"],
                    np.arange(detdict["nValuesExpected"]) / detdict["nValuesPerSecond"],
                    {"units": "s", "ancillary_variables": "elution_time_std_err"},
                ),
                "uts": (["uts"], [uts]),
            },
            attrs={},
        )
        valve = jsdata.get("annotations", {}).get("valcoPosition", None)
        if valve is not None:
            fvals["valve"] = valve
        vals[detname] = fvals

    dt = DataTree.from_dict(vals)
    dt.attrs = dict(original_metadata=metadata)
    return dt


def extract(
    *,
    fn: str,
    encoding: str,
    timezone: str,
    **kwargs: dict,
) -> DataTree:
    with open(fn, "r", encoding=encoding, errors="ignore") as infile:
        jsdata = json.load(infile)
    uts = dgutils.str_to_uts(timestamp=jsdata["runTimeStamp"], timezone=timezone)
    data = chromdata(jsdata, uts)
    trace = chromtrace(jsdata, uts)
    newdt = DataTree(data)
    for k, v in trace.items():
        newdt[k] = v
    return newdt



import pytest
import subprocess
import os
import json
from datatree import open_datatree
import numpy as np
from .utils import compare_datatrees


def test_yadg_version():
    command = ["yadg", "--version"]
    subprocess.run(command, check=True)


def test_yadg_process_without_schema(datadir):
    command = ["yadg", "process"]
    with pytest.raises(AssertionError, match="error: the following arguments"):
        try:
            subprocess.run(command, check=True, capture_output=True)
        except subprocess.CalledProcessError as err:
            assert False, err.stderr


def test_yadg_process_with_bad_schema(datadir):
    command = ["yadg", "process", "somefile.json"]
    with pytest.raises(AssertionError, match="Supplied dataschema filename"):
        try:
            subprocess.run(command, check=True, capture_output=True)
        except subprocess.CalledProcessError as err:
            assert False, err.stderr


def test_yadg_process_with_one_positional_arg(datadir):
    os.chdir(datadir)
    command = ["yadg", "process", "test_schema.json"]
    subprocess.run(command, check=True)
    assert os.path.exists("datagram.nc")


def test_yadg_process_with_two_positional_args(datadir):
    os.chdir(datadir)
    command = ["yadg", "process", "test_schema.json", "test.datagram.nc"]
    subprocess.run(command, check=True)
    assert os.path.exists("test.datagram.nc")


def test_yadg_update_without_subcommand(datadir):
    os.chdir(datadir)
    command = ["yadg", "update"]
    with pytest.raises(AssertionError, match="error: the following arguments"):
        try:
            subprocess.run(command, check=True, capture_output=True)
        except subprocess.CalledProcessError as err:
            assert False, err.stderr


def test_yadg_update_310(datadir):
    os.chdir(datadir)
    command = ["yadg", "update", "schema_3.1.0.json"]
    subprocess.run(command, check=True, capture_output=True)
    assert os.path.exists("schema_3.1.0.new.json")


def test_yadg_update_310_with_outfile(datadir):
    os.chdir(datadir)
    command = ["yadg", "update", "schema_3.1.0.json", "output.json"]
    subprocess.run(command, check=True, capture_output=True)
    assert os.path.exists("output.json")


def test_yadg_preset(datadir):
    os.chdir(datadir)
    command = ["yadg", "preset"]
    with pytest.raises(AssertionError, match="error: the following arguments"):
        try:
            subprocess.run(command, check=True, capture_output=True)
        except subprocess.CalledProcessError as err:
            assert False, err.stderr


def test_yadg_preset_with_preset_folder(datadir):
    os.chdir(datadir)
    command = ["yadg", "preset", "data_2.preset.json", "data_2"]
    subprocess.run(command, check=True, capture_output=True)
    assert os.path.exists("schema.json")


def test_yadg_preset_with_preset_folder_outfile(datadir):
    os.chdir(datadir)
    command = ["yadg", "preset", "data_2.preset.json", "data_2", "data_2.schema.json"]
    subprocess.run(command, check=True, capture_output=True)
    assert os.path.exists("data_2.schema.json")


def test_yadg_preset_with_preset_folder_p1(datadir):
    os.chdir(datadir)
    command = ["yadg", "preset", "data_2.preset.json", "data_2", "-p"]
    subprocess.run(command, check=True, capture_output=True)
    assert os.path.exists("datagram.nc")


def test_yadg_preset_with_preset_folder_p2(datadir):
    os.chdir(datadir)
    command = ["yadg", "preset", "data_2.preset.json", "data_2", "data_2.dg.nc", "-p"]
    subprocess.run(command, check=True, capture_output=True)
    assert os.path.exists("data_2.dg.nc")


def test_yadg_preset_with_preset_folder_p3(datadir):
    os.chdir(datadir)
    command = ["yadg", "preset", "data_2.preset.json", "data_2", "-p", "data_2.dg.nc"]
    subprocess.run(command, check=True, capture_output=True)
    assert os.path.exists("data_2.dg.nc")


def test_yadg_process_with_metadata(datadir):
    os.chdir(datadir)
    command = ["yadg", "process", "test_schema.yml"]
    subprocess.run(command, check=True)
    assert os.path.exists("datagram.nc")
    ret = open_datatree("datagram.nc", engine="h5netcdf")
    ref = open_datatree("datagram.nc.ref", engine="h5netcdf")
    compare_datatrees(ret, ref, thislevel=True, descend=True)


def test_yadg_preset_with_metadata(datadir):
    os.chdir(datadir)
    command = ["yadg", "preset", "-p", "data_2.preset.yaml", "data_2", "data_2.nc"]
    subprocess.run(command, check=True)
    assert os.path.exists("data_2.nc")
    ret = open_datatree("data_2.nc", engine="h5netcdf")
    ref = open_datatree("data_2.nc.ref", engine="h5netcdf")
    print(f"{ret.attrs=}")
    print(f"{ref.attrs=}")
    compare_datatrees(ret, ref, thislevel=True, descend=True)


@pytest.mark.parametrize(
    "packwith, suffix",
    [
        (None, "zip"),
        ("zip", "zip"),
        ("tar", "tar"),
        ("gztar", "tar.gz"),
        ("xztar", "tar.xz"),
        ("bztar", "tar.bz2"),
    ],
)
def test_yadg_preset_archive(packwith, suffix, datadir):
    os.chdir(datadir)
    command = ["yadg", "preset", "-pa", "data_2.preset.yaml", "data_2", "dg.nc"]
    if packwith is not None:
        command.append("--packwith")
        command.append(packwith)
    subprocess.run(command, check=True)
    assert os.path.exists("dg.nc")
    assert os.path.exists(f"dg.{suffix}")


def test_yadg_preset_roundtrip_uts(datadir):
    ts = {"nsteps": 1, "step": 0, "nrows": 20, "point": 19}
    ts["pars"] = {"uts": {"value": 1652254017.1712718}}
    os.chdir(datadir)
    command = ["yadg", "preset", "-p", "data_4.preset.json", "data_4", "data_4.nc"]
    subprocess.run(command, check=True)
    assert os.path.exists("data_4.nc")
    ret = open_datatree("data_4.nc", engine="h5netcdf")
    print(f"{ret=}")
    assert ret["worker"]["uts"].shape == (20,)
    np.testing.assert_almost_equal(ret["worker"]["uts"][-1], 1652254017.1712718)


@pytest.mark.parametrize(
    "filetype, infile",
    [
        ("eclab.mpr", "cp.mpr"),
        ("biologic-mpr", "cp.mpr"),
        ("agilent.ch", "agilent.CH"),
        ("fusion.json", "fusion.fusion-data"),
    ],
)
def test_yadg_extract_with_metadata(filetype, infile, datadir):
    os.chdir(datadir)
    command = ["yadg", "extract", filetype, infile, "test.nc"]
    subprocess.run(command, check=True)
    assert os.path.exists("test.nc")
    ret = open_datatree("test.nc", engine="h5netcdf")
    ref = open_datatree(f"{infile}.nc", engine="h5netcdf")
    # let's delete metadata we know will be wrong
    compare_datatrees(ret, ref, thislevel=True, descend=True)


@pytest.mark.parametrize(
    "filetype, infile, flag",
    [
        ("eclab.mpr", "cp.mpr", "-m"),
        ("biologic-mpr", "cp.mpr", "--meta-only"),
        ("agilent.ch", "agilent.CH", "-m"),
        ("fusion.json", "fusion.fusion-data", "--meta-only"),
    ],
)
def test_yadg_extract_meta_only(filetype, infile, flag, datadir):
    os.chdir(datadir)
    command = ["yadg", "extract", filetype, infile, flag]
    subprocess.run(command, check=True)
    outfile = infile.split(".")[0] + ".json"
    assert os.path.exists(outfile)
    with open(outfile, "r") as inp:
        ret = json.load(inp)
    print(f"{ret=}")
    for node in ret.values():
        for key in {"attrs", "coords", "dims", "data_vars"}:
            assert key in node.keys()


def test_yadg_preset_dataschema_compat(datadir):
    os.chdir(datadir)
    sfns = [fn for fn in os.listdir() if fn.endswith("yml") and fn.startswith("ds")]
    ncs = []
    for fn in sfns:
        command = ["yadg", "preset", "-p", fn, "ds_compat", fn.replace("yml", "nc")]
        subprocess.run(command, check=True)
        ncs.append((fn, open_datatree(fn.replace("yml", "nc"), engine="h5netcdf")))
    _, ref = ncs[0]
    for name, ret in ncs[1:]:
        try:
            compare_datatrees(ret, ref)
        except AssertionError as e:
            e.args = (e.args[0] + f"\nFailed on file {name!r}.\n",)
            raise e


@pytest.mark.parametrize(
    "filetype, infile, locale",
    [
        ("eclab.mpt", "mb.issue_95.de.mpt", "de_DE"),
    ],
)
def test_yadg_extract_locale(filetype, infile, locale, datadir):
    os.chdir(datadir)
    command = ["yadg", "extract", "--locale", locale, filetype, infile, "test.nc"]
    subprocess.run(command, check=True)
    assert os.path.exists("test.nc")
    ret = open_datatree("test.nc", engine="h5netcdf")
    ref = open_datatree(f"{infile}.nc", engine="h5netcdf")
    compare_datatrees(ret, ref, thislevel=True, descend=True)
